from __future__ import annotations

import importlib
import itertools
import pathlib
from typing import TYPE_CHECKING

from attrs import Factory, converters, define, field, fields, validators
from typing_extensions import deprecated

from eta_utility import deep_mapping_update, dict_pop_any, get_logger, json_import
from eta_utility.util import toml_import, yaml_import

if TYPE_CHECKING:
    from collections.abc import Mapping
    from typing import Any

    from attrs import Attribute
    from stable_baselines3.common.base_class import BaseAlgorithm, BasePolicy
    from stable_baselines3.common.vec_env import DummyVecEnv

    from eta_utility.eta_x.envs import BaseEnv
    from eta_utility.type_hints import Path


log = get_logger("eta_x")


def _path_converter(path: Path) -> pathlib.Path:
    """Convert value to a class."""
    return pathlib.Path(path) if not isinstance(path, pathlib.Path) else path


def _get_class(instance: ConfigOptSetup, attrib: Attribute, new_value: str | None) -> str | None:
    """Find module and class name and import the specified class."""
    if new_value is not None:
        module, cls_name = new_value.rsplit(".", 1)
        try:
            cls = getattr(importlib.import_module(module), cls_name)
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                f"Could not find module '{e.name}'. While importing class '{cls_name}' from '{attrib.name}' value."
            ) from e
        except AttributeError as e:
            raise AttributeError(
                f"Could not find class '{cls_name}' in module '{module}'. "
                f"While importing class '{cls_name}' from '{attrib.name}' value."
            ) from e

        cls_attr_name = f"{attrib.name.rsplit('_', 1)[0]}_class"
        setattr(instance, cls_attr_name, cls)

    return new_value


@define(frozen=False, kw_only=True)
class ConfigOpt:
    """Configuration for the optimization, which can be loaded from a JSON file."""

    #: Name of the configuration used for the series of run.
    config_name: str = field(validator=validators.instance_of(str))

    #: Root path for the optimization run (scenarios and results are relative to this).
    path_root: pathlib.Path = field(converter=_path_converter)
    #: Relative path to the results folder.
    relpath_results: str = field(validator=validators.instance_of(str))
    #: relative path to the scenarios folder (default: None).
    relpath_scenarios: str | None = field(validator=validators.optional(validators.instance_of(str)), default=None)
    #: Path to the results folder.
    path_results: pathlib.Path = field(init=False, converter=_path_converter)
    #: Path to the scenarios folder (default: None).
    path_scenarios: pathlib.Path | None = field(
        init=False, converter=converters.optional(_path_converter), default=None
    )

    #: Optimization run setup.
    setup: ConfigOptSetup = field()
    #: Optimization run settings.
    settings: ConfigOptSettings = field()

    def __attrs_post_init__(self) -> None:
        object.__setattr__(self, "path_results", self.path_root / self.relpath_results)

        if self.relpath_scenarios is not None:
            object.__setattr__(self, "path_scenarios", self.path_root / self.relpath_scenarios)

    @classmethod
    @deprecated("Use `ConfigOpt.from_config_file()` instead.")
    def from_json(cls, file: Path, path_root: Path, overwrite: Mapping[str, Any] | None = None) -> ConfigOpt:
        return cls.from_config_file(file=file, path_root=path_root, overwrite=overwrite)

    @classmethod
    def from_config_file(cls, file: Path, path_root: Path, overwrite: Mapping[str, Any] | None = None) -> ConfigOpt:
        """Load configuration from JSON/TOML/YAML file, which consists of the following sections:

        - **paths**: In this section, the (relative) file paths for results and scenarios are specified. The paths
          are deserialized directly into the :class:`ConfigOpt` object.
        - **setup**: This section specifies which classes and utilities should be used for optimization. The setup
          configuration is deserialized into the :class:`ConfigOptSetup` object.
        - **settings**: The settings section contains basic parameters for the optimization, it is deserialized
          into a :class:`ConfigOptSettings` object.
        - **environment_specific**: The environment section contains keyword arguments for the environment.
          This section must contain values for the arguments of the environment, the expected values are therefore
          different depending on the environment and not fully documented here.
        - **agent_specific**: The agent section contains keyword arguments for the control algorithm (agent).
          This section must contain values for the arguments of the agent, the expected values are therefore
          different depending on the agent and not fully documented here.

        :param file: Path to the configuration file.
        :param overwrite: Config parameters to overwrite.
        :return: ConfigOpt object.
        """
        _path_root: pathlib.Path = pathlib.Path(path_root)

        config = cls._load_config_file(file)

        if overwrite is not None:
            config = dict(deep_mapping_update(config, overwrite))

        # Ensure all required sections are present in configuration
        for section in ("setup", "settings", "paths"):
            if section not in config:
                raise ValueError(f"The section '{section}' is not present in configuration file {file}.")

        return ConfigOpt.from_dict(config, file, _path_root)

    @staticmethod
    def _load_config_file(file: Path) -> dict[str, Any]:
        """Load configuration file from JSON, TOML, or YAML file.
        The read file is expected to contain a dictionary of configuration options.

        :param file: Path to the configuration file.
        :return: Dictionary of configuration options.
        """
        possible_extensions = {".json": json_import, ".toml": toml_import, ".yml": yaml_import, ".yaml": yaml_import}
        file_path = pathlib.Path(file)

        for extension in possible_extensions:
            _file_path: pathlib.Path = file_path.with_suffix(extension)
            if _file_path.exists():
                result = possible_extensions[extension](_file_path)
                break
        else:
            raise FileNotFoundError(f"Config file not found: {file}")

        if not isinstance(result, dict):
            raise TypeError(f"Config file {file} must define a dictionary of options.")

        return result

    @classmethod
    def from_dict(cls, config: dict[str, Any], file: Path, path_root: pathlib.Path) -> ConfigOpt:
        """Build a ConfigOpt object from a dictionary of configuration options.

        :param config: Dictionary of configuration options.
        :param file: Path to the configuration file.
        :param path_root: Root path for the optimization configuration run.
        :return: ConfigOpt object.
        """

        def _pop_dict(dikt: dict[str, Any], key: str) -> dict[str, Any]:
            val = dikt.pop(key)
            if not isinstance(val, dict):
                raise TypeError(f"'{key}' section must be a dictionary of settings.")
            return val

        if "environment_specific" not in config:
            config["environment_specific"] = {}
            log.info("Section 'environment_specific' not present in configuration, assuming it is empty.")

        if "agent_specific" not in config:
            config["agent_specific"] = {}
            log.info("Section 'agent_specific' not present in configuration, assuming it is empty.")

        # Load values from paths section
        errors = False
        paths = _pop_dict(config, "paths")

        if "relpath_results" not in paths:
            log.error("'relpath_results' is required and could not be found in section 'paths' of the configuration.")
            errors = True
        relpath_results = paths.pop("relpath_results", None)
        relpath_scenarios = paths.pop("relpath_scenarios", None)

        # Setup section
        _setup = _pop_dict(config, "setup")
        try:
            setup = ConfigOptSetup.from_dict(_setup)
        except ValueError as e:
            log.error(e)
            errors = True

        # Settings section
        settings_raw: dict[str, dict[str, Any]] = {}
        settings_raw["settings"] = _pop_dict(config, "settings")
        settings_raw["environment_specific"] = _pop_dict(config, "environment_specific")

        if "interaction_env_specific" in config:
            settings_raw["interaction_env_specific"] = _pop_dict(config, "interaction_env_specific")
        elif "interaction_environment_specific" in config:
            settings_raw["interaction_env_specific"] = _pop_dict(config, "interaction_environment_specific")

        settings_raw["agent_specific"] = _pop_dict(config, "agent_specific")

        try:
            settings = ConfigOptSettings.from_dict(settings_raw)
        except ValueError as e:
            log.error(e)
            errors = True
        # Log configuration values which were not recognized.
        for name in config:
            log.warning(
                f"Specified configuration value '{name}' in the setup section of the configuration was not "
                f"recognized and is ignored."
            )

        if errors:
            raise ValueError(
                "Not all required values were found in setup section (see log). Could not load config file."
            )

        return cls(
            config_name=str(file),
            path_root=path_root,
            relpath_results=relpath_results,
            relpath_scenarios=relpath_scenarios,
            setup=setup,
            settings=settings,
        )

    def __getitem__(self, name: str) -> Any:
        return getattr(self, name)

    def __setitem__(self, name: str, value: Any) -> None:
        if not hasattr(self, name):
            raise KeyError(f"The key {name} does not exist - it cannot be set.")
        setattr(self, name, value)


@define(frozen=False, kw_only=True)
class ConfigOptSetup:
    """Configuration options as specified in the "setup" section of the configuration file."""

    #: Import description string for the agent class.
    agent_import: str = field(on_setattr=_get_class)
    #: Agent class (automatically determined from agent_import).
    agent_class: type[BaseAlgorithm] = field(init=False)
    #: Import description string for the environment class.
    environment_import: str = field(on_setattr=_get_class)
    #: Imported Environment class (automatically determined from environment_import).
    environment_class: type[BaseEnv] = field(init=False)
    #: Import description string for the interaction environment (default: None).
    interaction_env_import: str | None = field(default=None, on_setattr=_get_class)
    #: Interaction environment class (default: None) (automatically determined from interaction_env_import).
    interaction_env_class: type[BaseEnv] | None = field(init=False, default=None)

    #: Import description string for the environment vectorizer
    #: (default: stable_baselines3.common.vec_env.dummy_vec_env.DummyVecEnv).
    vectorizer_import: str = field(
        default="stable_baselines3.common.vec_env.dummy_vec_env.DummyVecEnv",
        on_setattr=_get_class,
        converter=converters.default_if_none(  # type: ignore
            "stable_baselines3.common.vec_env.dummy_vec_env.DummyVecEnv"
        ),
    )  # mypy currently does not recognize converters.default_if_none
    #: Environment vectorizer class  (automatically determined from vectorizer_import).
    vectorizer_class: type[DummyVecEnv] = field(init=False)
    #: Import description string for the policy class (default: eta_utility.eta_x.agents.common.NoPolicy).
    policy_import: str = field(
        default="eta_utility.eta_x.common.NoPolicy",
        on_setattr=_get_class,
        converter=converters.default_if_none("eta_utility.eta_x.common.NoPolicy"),  # type: ignore
    )  # mypy currently does not recognize converters.default_if_none
    #: Policy class (automatically determined from policy_import).
    policy_class: type[BasePolicy] = field(init=False)

    #: Flag which is true if the environment should be wrapped for monitoring (default: False).
    monitor_wrapper: bool = field(default=False, converter=bool)
    #: Flag which is true if the observations should be normalized (default: False).
    norm_wrapper_obs: bool = field(default=False, converter=bool)
    #: Flag which is true if the rewards should be normalized (default: False).
    norm_wrapper_reward: bool = field(default=False, converter=bool)
    #: Flag to enable tensorboard logging (default: False).
    tensorboard_log: bool = field(default=False, converter=bool)

    def __attrs_post_init__(self) -> None:
        _fields = fields(ConfigOptSetup)
        _get_class(self, _fields.agent_import, self.agent_import)
        _get_class(self, _fields.environment_import, self.environment_import)
        _get_class(self, _fields.interaction_env_import, self.interaction_env_import)
        _get_class(self, _fields.vectorizer_import, self.vectorizer_import)
        _get_class(self, _fields.policy_import, self.policy_import)

    @classmethod
    def from_dict(cls, dikt: dict[str, Any]) -> ConfigOptSetup:
        errors = []

        def get_import(name: str, required: bool = False) -> str | Any:
            """Get import string or combination of package and class name from dictionary.
            :param name: Name of the configuration value.
            :param required: Flag to determine if the value is required.
            """
            nonlocal errors, dikt
            import_value = dikt.pop(f"{name}_import", None)
            package_value = dikt.pop(f"{name}_package", None)
            class_value = dikt.pop(f"{name}_class", None)
            # Check import
            if import_value is not None:
                return import_value

            # Check package and class
            if package_value is not None and class_value is not None:
                return f"{package_value}.{class_value}"

            # If only one of package and class is specified, raise error
            if (package_value is None) ^ (class_value is None):
                msg = f"Only one of '{name}_package' and '{name}_class' is specified."
                log.info(msg)

            # Raise error if required value is missing
            if required:
                msg = f"'{name}_import' or both of '{name}_package' and '{name}_class' parameters must be specified."
                log.error(msg)
                errors.append(name)
            return None

        agent_import = get_import("agent", required=True)
        environment_import = get_import("environment", required=True)

        interaction_env_import = get_import("interaction_env")
        vectorizer_import = get_import("vectorizer")
        policy_import = get_import("policy")

        monitor_wrapper = dikt.pop("monitor_wrapper", None)
        norm_wrapper_obs = dikt.pop("norm_wrapper_obs", None)
        norm_wrapper_reward = dikt.pop("norm_wrapper_reward", None)
        tensorboard_log = dikt.pop("tensorboard_log", None)

        # Log configuration values which were not recognized.
        if dikt:
            msg = "Following values were not recognized in the config setup section and are ignored: "
            msg += ", ".join(dikt.keys())
            log.warning(msg)

        if errors:
            msg = "Not all required values were found in setup section (see log). Could not load config file. "
            msg += f"Missing values: {', '.join(errors)}"
            raise ValueError(msg)

        return ConfigOptSetup(
            agent_import=agent_import,
            environment_import=environment_import,
            interaction_env_import=interaction_env_import,
            vectorizer_import=vectorizer_import,
            policy_import=policy_import,
            monitor_wrapper=monitor_wrapper,
            norm_wrapper_obs=norm_wrapper_obs,
            norm_wrapper_reward=norm_wrapper_reward,
            tensorboard_log=tensorboard_log,
        )

    def __getitem__(self, name: str) -> Any:
        return getattr(self, name)

    def __setitem__(self, name: str, value: Any) -> None:
        if not hasattr(self, name):
            raise KeyError(f"The key {name} does not exist - it cannot be set.")
        setattr(self, name, value)


def _env_defaults(instance: ConfigOptSettings, attrib: Attribute, new_value: dict[str, Any] | None) -> dict[str, Any]:
    """Set default values for the environment settings."""
    _new_value = {} if new_value is None else new_value

    _new_value.setdefault("verbose", instance.verbose)
    _new_value.setdefault("sampling_time", instance.sampling_time)
    _new_value.setdefault("episode_duration", instance.episode_duration)

    if instance.sim_steps_per_sample is not None:
        _new_value.setdefault("sim_steps_per_sample", instance.sim_steps_per_sample)

    return _new_value


def _agent_defaults(instance: ConfigOptSettings, attrib: Attribute, new_value: dict[str, Any] | None) -> dict[str, Any]:
    """Set default values for the environment settings."""
    _new_value = {} if new_value is None else new_value

    _new_value.setdefault("seed", instance.seed)
    _new_value.setdefault("verbose", instance.verbose)

    return _new_value


@define(frozen=False, kw_only=True)
class ConfigOptSettings:
    #: Seed for random sampling (default: None).
    seed: int | None = field(default=None, converter=converters.optional(int))
    #: Logging verbosity of the framework (default: 2).
    verbose: int = field(
        default=2,
        converter=converters.pipe(converters.default_if_none(2), int),  # type: ignore
    )  # mypy currently does not recognize converters.default_if_none
    #: Number of vectorized environments to instantiate (if not using DummyVecEnv) (default: 1).
    n_environments: int = field(
        default=1,
        converter=converters.pipe(converters.default_if_none(1), int),  # type: ignore
    )  # mypy currently does not recognize converters.default_if_none

    #: Number of episodes to execute when the agent is playing (default: None).
    n_episodes_play: int | None = field(default=None, converter=converters.optional(int))
    #: Number of episodes to execute when the agent is learning (default: None).
    n_episodes_learn: int | None = field(default=None, converter=converters.optional(int))
    #: Flag to determine whether the interaction env is used or not (default: False).
    interact_with_env: bool = field(
        default=False,
        converter=converters.pipe(converters.default_if_none(False), bool),  # type: ignore
    )  # mypy currently does not recognize converters.default_if_none
    #: How often to save the model during training (default: 10 - after every ten episodes).
    save_model_every_x_episodes: int = field(
        default=10,
        converter=converters.pipe(converters.default_if_none(1), int),  # type: ignore
    )  # mypy currently does not recognize converters.default_if_none
    #: How many episodes to pass between each render call (default: 10 - after every ten episodes).
    plot_interval: int = field(
        default=10,
        converter=converters.pipe(converters.default_if_none(1), int),  # type: ignore
    )  # mypy currently does not recognize converters.default_if_none

    #: Duration of an episode in seconds (can be a float value).
    episode_duration: float = field(converter=float)
    #: Duration between time samples in seconds (can be a float value).
    sampling_time: float = field(converter=float)
    #: Simulation steps for every sample.
    sim_steps_per_sample: int | None = field(default=None, converter=converters.optional(int))

    #: Multiplier for scaling the agent actions before passing them to the environment
    #: (especially useful with interaction environments) (default: None).
    scale_actions: float | None = field(default=None, converter=converters.optional(float))
    #: Number of digits to round actions to before passing them to the environment
    #: (especially useful with interaction environments) (default: None).
    round_actions: int | None = field(default=None, converter=converters.optional(int))

    #: Settings dictionary for the environment.
    environment: dict[str, Any] = field(
        default=Factory(dict),
        converter=converters.default_if_none(Factory(dict)),  # type: ignore
        on_setattr=_env_defaults,
    )  # mypy currently does not recognize converters.default_if_none
    #: Settings dictionary for the interaction environment (default: None).
    interaction_env: dict[str, Any] | None = field(default=None, on_setattr=_env_defaults)
    #: Settings dictionary for the agent.
    agent: dict[str, Any] = field(
        default=Factory(dict),
        converter=converters.default_if_none(Factory(dict)),  # type: ignore
        # mypy currently does not recognize converters.default_if_none
        on_setattr=_agent_defaults,
    )

    #: Flag which is true if the log output should be written to a file
    log_to_file: bool = field(
        default=True,
        converter=converters.pipe(converters.default_if_none(False), bool),  # type: ignore
    )

    def __attrs_post_init__(self) -> None:
        _fields = fields(ConfigOptSettings)
        _env_defaults(self, _fields.environment, self.environment)
        _agent_defaults(self, _fields.agent, self.agent)

        # Set standards for interaction env settings or copy settings from environment
        if self.interaction_env is not None:
            _env_defaults(self, _fields.interaction_env, self.interaction_env)
        elif self.interact_with_env is True and self.interaction_env is None:
            log.warning(
                "Interaction with an environment has been requested, but no section 'interaction_env_specific' "
                "found in settings. Re-using 'environment_specific' section."
            )
            self.interaction_env = self.environment

        if self.n_episodes_play is None and self.n_episodes_learn is None:
            raise ValueError("At least one of 'n_episodes_play' or 'n_episodes_learn' must be specified in settings.")

    @classmethod
    def from_dict(cls, dikt: dict[str, dict[str, Any]]) -> ConfigOptSettings:
        errors = False

        # Read general settings dictionary
        if "settings" not in dikt:
            raise ValueError("Settings section not found in configuration. Cannot import config file.")
        settings = dikt.pop("settings")

        if "seed" not in settings:
            log.info("'seed' not specified in settings, using default value 'None'")
        seed = settings.pop("seed", None)

        if "verbose" not in settings and "verbosity" not in settings:
            log.info("'verbose' or 'verbosity' not specified in settings, using default value '2'")
        verbose = dict_pop_any(settings, "verbose", "verbosity", fail=False, default=None)

        if "n_environments" not in settings:
            log.info("'n_environments' not specified in settings, using default value '1'")
        n_environments = settings.pop("n_environments", None)

        if "n_episodes_play" not in settings and "n_episodes_learn" not in settings:
            log.error("Neither 'n_episodes_play' nor 'n_episodes_learn' is specified in settings.")
            errors = True
        n_epsiodes_play = settings.pop("n_episodes_play", None)
        n_episodes_learn = settings.pop("n_episodes_learn", None)

        interact_with_env = settings.pop("interact_with_env", False)
        save_model_every_x_episodes = settings.pop("save_model_every_x_episodes", None)
        plot_interval = settings.pop("plot_interval", None)

        if "episode_duration" not in settings:
            log.error("'episode_duration' is not specified in settings.")
            errors = True
        episode_duration = settings.pop("episode_duration", None)

        if "sampling_time" not in settings:
            log.error("'sampling_time' is not specified in settings.")
            errors = True
        sampling_time = settings.pop("sampling_time", None)

        sim_steps_per_sample = settings.pop("sim_steps_per_sample", None)
        scale_actions = dict_pop_any(settings, "scale_interaction_actions", "scale_actions", fail=False, default=None)
        round_actions = dict_pop_any(settings, "round_interaction_actions", "round_actions", fail=False, default=None)

        if "environment_specific" not in dikt:
            log.error("'environment_specific' section not defined in settings.")
            errors = True
        environment = dikt.pop("environment_specific", None)

        if "agent_specific" not in dikt:
            log.error("'agent_specific' section not defined in settings.")
            errors = True
        agent = dikt.pop("agent_specific", None)

        interaction_env = dict_pop_any(
            dikt, "interaction_env_specific", "interaction_environment_specific", fail=False, default=None
        )

        log_to_file = settings.pop("log_to_file", False)

        # Log configuration values which were not recognized.
        for name in itertools.chain(settings, dikt):
            log.warning(
                f"Specified configuration value '{name}' in the settings section of the configuration "
                f"was not recognized and is ignored."
            )

        if errors:
            raise ValueError("Not all required values were found in settings (see log). Could not load config file.")

        return cls(
            seed=seed,
            verbose=verbose,
            n_environments=n_environments,
            n_episodes_play=n_epsiodes_play,
            n_episodes_learn=n_episodes_learn,
            interact_with_env=interact_with_env,
            save_model_every_x_episodes=save_model_every_x_episodes,
            plot_interval=plot_interval,
            episode_duration=episode_duration,
            sampling_time=sampling_time,
            sim_steps_per_sample=sim_steps_per_sample,
            scale_actions=scale_actions,
            round_actions=round_actions,
            environment=environment,
            agent=agent,
            interaction_env=interaction_env,
            log_to_file=log_to_file,
        )

    def __getitem__(self, name: str) -> Any:
        return getattr(self, name)

    def __setitem__(self, name: str, value: Any) -> None:
        if not hasattr(self, name):
            raise KeyError(f"The key {name} does not exist - it cannot be set.")
        setattr(self, name, value)


@define(frozen=True, kw_only=True)
class ConfigOptRun:
    """Configuration for an optimization run, including the series and run names descriptions and paths
    for the run.
    """

    #: Name of the series of optimization runs.
    series: str = field(validator=validators.instance_of(str))
    #: Name of an optimization run.
    name: str = field(validator=validators.instance_of(str))
    #: Description of an optimization run.
    description: str = field(
        converter=converters.default_if_none(""),  # type: ignore
        validator=validators.instance_of(str),
    )
    #: Root path of the framework run.
    path_root: pathlib.Path = field(converter=_path_converter)
    #: Path to results of the optimization run.
    path_results: pathlib.Path = field(converter=_path_converter)
    #: Path to scenarios used for the optimization run.
    path_scenarios: pathlib.Path | None = field(default=None, converter=converters.optional(_path_converter))
    #: Path for the results of the series of optimization runs.
    path_series_results: pathlib.Path = field(init=False, converter=_path_converter)
    #: Path to the model of the optimization run.
    path_run_model: pathlib.Path = field(init=False, converter=_path_converter)
    #: Path to information about the optimization run.
    path_run_info: pathlib.Path = field(init=False, converter=_path_converter)
    #: Path to the monitoring information about the optimization run.
    path_run_monitor: pathlib.Path = field(init=False, converter=_path_converter)
    #: Path to the normalization wrapper information.
    path_vec_normalize: pathlib.Path = field(init=False, converter=_path_converter)
    #: Path to the neural network architecture file.
    path_net_arch: pathlib.Path = field(init=False, converter=_path_converter)
    #: Path to the log output file.
    path_log_output: pathlib.Path = field(init=False, converter=_path_converter)

    # Information about the environments
    #: Version of the main environment.
    env_version: str | None = field(
        init=False, default=None, validator=validators.optional(validators.instance_of(str))
    )
    #: Description of the main environment.
    env_description: str | None = field(
        init=False, default=None, validator=validators.optional(validators.instance_of(str))
    )

    #: Version of the secondary environment (interaction_env).
    interaction_env_version: str | None = field(
        init=False, default=None, validator=validators.optional(validators.instance_of(str))
    )
    #: Description of the secondary environment (interaction_env).
    interaction_env_description: str | None = field(
        init=False, default=None, validator=validators.optional(validators.instance_of(str))
    )

    def __attrs_post_init__(self) -> None:
        """Add default values to the derived paths."""
        object.__setattr__(self, "path_series_results", self.path_results / self.series)
        object.__setattr__(self, "path_run_model", self.path_series_results / f"{self.name}_model.zip")
        object.__setattr__(self, "path_run_info", self.path_series_results / f"{self.name}_info.json")
        object.__setattr__(self, "path_run_monitor", self.path_series_results / f"{self.name}_monitor.csv")
        object.__setattr__(self, "path_vec_normalize", self.path_series_results / "vec_normalize.pkl")
        object.__setattr__(self, "path_net_arch", self.path_series_results / "net_arch.txt")
        object.__setattr__(self, "path_log_output", self.path_series_results / f"{self.name}_log_output.log")

    def create_results_folders(self) -> None:
        """Create the results folders for an optimization run (or check if they already exist)."""
        if not self.path_results.is_dir():
            for p in reversed(self.path_results.parents):
                if not p.is_dir():
                    p.mkdir()
                    log.info(f"Directory created: \n\t {p}")
            self.path_results.mkdir()
            log.info(f"Directory created: \n\t {self.path_results}")

        if not self.path_series_results.is_dir():
            log.debug("Path for result series doesn't exist on your OS. Trying to create directories.")
            self.path_series_results.mkdir()
            log.info(f"Directory created: \n\t {self.path_series_results}")

    def set_env_info(self, env: type[BaseEnv]) -> None:
        """Set the environment information of the optimization run to represent the given environment.
        The information will default to None if this is never called.

        :param env: The environment whose description should be used.
        """
        version, description = env.get_info()
        object.__setattr__(self, "env_version", version)
        object.__setattr__(self, "env_description", description)

    def set_interaction_env_info(self, env: type[BaseEnv]) -> None:
        """Set the interaction environment information of the optimization run to represent the given environment.
        The information will default to None if this is never called.

        :param env: The environment whose description should be used.
        """
        version, description = env.get_info()
        object.__setattr__(self, "interaction_env_version", version)
        object.__setattr__(self, "interaction_env_description", description)

    @property
    def paths(self) -> dict[str, pathlib.Path]:
        """Dictionary of all paths for the optimization run. This is for easier access and contains all
        paths as mentioned above."""
        paths = {
            "path_root": self.path_root,
            "path_results": self.path_results,
            "path_series_results": self.path_series_results,
            "path_run_model": self.path_run_model,
            "path_run_info": self.path_run_info,
            "path_run_monitor": self.path_run_monitor,
            "path_vec_normalize": self.path_vec_normalize,
            "path_log_output": self.path_log_output,
        }
        if self.path_scenarios is not None:
            paths["path_scenarios"] = self.path_scenarios

        return paths



from __future__ import annotations

import datetime
from os import PathLike
from typing import Union

import numpy as np
from cryptography.hazmat.primitives.asymmetric import (
    dh,
    dsa,
    ec,
    ed448,
    ed25519,
    rsa,
    x448,
    x25519,
)

# Other custom types:
Path = Union[str, PathLike]
Number = Union[float, int, np.floating, np.signedinteger, np.unsignedinteger]
TimeStep = Union[int, float, datetime.timedelta]

PrivateKey = Union[
    dh.DHPrivateKey,
    ed25519.Ed25519PrivateKey,
    ed448.Ed448PrivateKey,
    rsa.RSAPrivateKey,
    dsa.DSAPrivateKey,
    ec.EllipticCurvePrivateKey,
    x25519.X25519PrivateKey,
    x448.X448PrivateKey,
]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar

import pandas as pd
from wetterdienst import Settings
from wetterdienst.provider.dwd.mosmix.api import DwdMosmixRequest
from wetterdienst.provider.dwd.observation.api import DwdObservationRequest

if TYPE_CHECKING:
    from collections.abc import Mapping
    from typing import Any

    from wetterdienst.core.timeseries.result import StationsResult

    from eta_utility.type_hints import Nodes, TimeStep

from datetime import datetime, timedelta

from eta_utility import get_logger
from eta_utility.connectors.node import (
    NodeWetterdienst,
    NodeWetterdienstObservation,
    NodeWetterdienstPrediction,
)

from .base_classes import SeriesConnection, SubscriptionHandler

log = get_logger("connectors.wetterdienst")

NW = TypeVar("NW", bound=NodeWetterdienst)


class WetterdienstConnection(Generic[NW], SeriesConnection[NW], ABC):
    """
    The WetterdienstConnection class is a connector to the Wetterdienst API for retrieving weather data.
    This class is an abstract base class and should not be used directly. Instead, use the subclasses
    :class:`WetterdienstObservationConnection` and :class:`WetterdienstPredictionConnection`.

    :param url: The base URL of the Wetterdienst API
    :param nodes: Nodes to select in connection
    :param settings: Wetterdienst settings object
    """

    def __init__(
        self,
        *,
        nodes: Nodes[NW] | None = None,
        settings: Settings | None = None,
        **kwargs: Any,
    ) -> None:
        self.settings = Settings(settings=settings)
        self.settings.ts_skip_empty = True
        self.settings.ts_si_units = False
        self.settings.ts_humanize = True
        super().__init__("https://opendata.dwd.de/", nodes=nodes)  # dummy url

    @classmethod
    def _from_node(cls, node: NW, **kwargs: Any) -> WetterdienstConnection:
        """Initialize the connection object from an wetterdienst protocol node object

        :param node: Node to initialize from
        :param kwargs: Extra keyword arguments
        """
        settings = kwargs.get("settings")
        return super()._from_node(node, settings=settings)

    @abstractmethod
    def read_series(
        self,
        from_time: datetime,
        to_time: datetime,
        nodes: Nodes[NW] | None = None,
        interval: TimeStep = 60,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """Abstract base method for read_series(). Is fully implemented in
        :func:`~wetterdienst.WetterdienstObservationConnection.read_series` and
        :func:`~wetterdienst.WetterdienstPredictionConnection.read_series`.

        :param nodes: List of nodes to read values from.
        :param from_time: Starting time to begin reading (included in output).
        :param to_time: Time to stop reading at (not included in output).
        :param interval: interval between time steps. It is interpreted as seconds if given as integer.
        :param kwargs: additional argument list, to be defined by subclasses.
        :return: pandas.DataFrame containing the data read from the connection.
        """
        if from_time.tzinfo != to_time.tzinfo:
            log.warning(
                f"Timezone of from_time and to_time are different. Using from_time timezone: {from_time.tzinfo}"
            )

    def read(self, nodes: Nodes[NW] | None = None) -> pd.DataFrame:
        """
        .. warning::
            Cannot read single values from the Wetterdienst API. Use read_series instead

        :param nodes: List of nodes to read values from
        :return: Pandas DataFrame containing the data read from the connection
        """
        raise NotImplementedError("Cannot read single values from the Wetterdienst API. Use read_series instead")

    def write(self, values: Mapping[NW, Any], time_interval: timedelta | None = None) -> None:
        """
        .. warning::
            Cannot write to the Wetterdienst API.

        :param values: Dictionary of nodes and data to write. {node: value}
        :param time_interval: Interval between datapoints, default 1s
        """
        raise NotImplementedError("Cannot write to the Wetterdienst API.")

    def subscribe(self, handler: SubscriptionHandler, nodes: Nodes[NW] | None = None, interval: TimeStep = 1) -> None:
        """Subscribe to nodes and call handler when new data is available. This will return only the
        last available values.

        :param handler: SubscriptionHandler object with a push method that accepts node, value pairs
        :param interval: interval for receiving new data. It is interpreted as seconds when given as an integer.
        :param nodes: identifiers for the nodes to subscribe to
        """
        raise NotImplementedError("Cannot subscribe to data from the Wetterdienst API.")

    def subscribe_series(
        self,
        handler: SubscriptionHandler,
        req_interval: TimeStep,
        offset: TimeStep | None = None,
        nodes: Nodes[NW] | None = None,
        interval: TimeStep = 1,
        data_interval: TimeStep = 1,
        **kwargs: Any,
    ) -> None:
        """
        .. warning::
            Not implemented: Cannot subscribe to data from the Wetterdienst API.

        :param handler: SubscriptionHandler object with a push method that accepts node, value pairs
        :param req_interval: Duration covered by requested data (time interval). Interpreted as seconds if given as int
        :param offset: Offset from datetime.now from which to start requesting data (time interval).
                       Interpreted as seconds if given as int. Use negative values to go to past timestamps.
        :param data_interval: Time interval between values in returned data. Interpreted as seconds if given as int.
        :param interval: interval (between requests) for receiving new data.
                         It it interpreted as seconds when given as an integer.
        :param nodes: identifiers for the nodes to subscribe to
        """
        raise NotImplementedError("Cannot subscribe to data from the Wetterdienst API.")

    def close_sub(self) -> None:
        """
        .. warning::
            Not implemented: Cannot subscribe to data from the the Wetterdienst API.
        """
        raise NotImplementedError("Cannot subscribe to data from the the Wetterdienst API.")

    def retrieve_stations(self, node: NodeWetterdienst, request: DwdObservationRequest) -> pd.DataFrame:
        """
        Retrieve stations from the Wetterdienst API and return the values as a pandas DataFrame
        Stations are filtered by the node's station_id or latlon and number_of_stations

        :param node: Node to retrieve stations for
        :param request: Wetterdienst request object, containing the station data
        """
        # Retrieve stations. If station_id is provided, use it, otherwise use latlon to get nearest stations
        stations: StationsResult
        if node.station_id is not None:
            stations = request.filter_by_station_id(node.station_id)
        else:
            stations = request.filter_by_rank(node.latlon, rank=node.number_of_stations)

        # Convert to pandas and pivot values so date is the index and
        # node names combined with the station_id are the columns
        result_df: pd.DataFrame = stations.values.all().df.to_pandas()  # noqa: PD011 (stations is not a dataframe)
        result_df = result_df.pivot_table(values="value", columns=("parameter", "station_id"), index="date")

        # Rename the columns to the node names
        result_df = result_df.rename({node.parameter.lower(): node.name}, axis="columns")
        return result_df.rename_axis(("Name", "station_id"), axis="columns")


class WetterdienstObservationConnection(
    WetterdienstConnection[NodeWetterdienstObservation], protocol="wetterdienst_observation"
):
    """
    The WetterdienstObservationConnection class is a connector to the Wetterdienst API
    for retrieving weather observation data. Data can only be read with
    :func:`~wetterdienst.WetterdienstObservationConnection.read_series`.
    """

    def read_series(
        self,
        from_time: datetime,
        to_time: datetime,
        nodes: Nodes[NodeWetterdienstObservation] | None = None,
        interval: TimeStep = 60,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """
        Read weather observation data from the Wetterdienst API for the given nodes and time interval

        :param from_time: Start time for the data retrieval
        :param to_time: End time for the data retrieval
        :param nodes: Nodes to read data from
        :param interval: Time interval between data points in seconds
        :return: Pandas DataFrame containing the data read from the connection
        """

        super().read_series(from_time, to_time, nodes, interval)
        nodes = self._validate_nodes(nodes)
        interval = interval if isinstance(interval, timedelta) else timedelta(seconds=interval)

        def _read_node(node: NodeWetterdienstObservation) -> pd.Dataframe:
            # Get the resolution for the node from the interval
            resolution = NodeWetterdienstObservation.convert_interval_to_resolution(node.interval)
            # Create a request object for the node
            request: DwdObservationRequest = DwdObservationRequest(
                parameter=node.parameter,
                resolution=resolution,
                start_date=from_time,
                end_date=to_time,
                settings=self.settings,
            )
            return self.retrieve_stations(node, request)

        # We can't use a ThreadPoolExecutor here, as the Wetterdienst library uses asyncio.
        # As a result, we have to call the _read_node method directly, which causes type errors.
        results = []
        for node in nodes:
            results.append(_read_node(node))
        result = pd.concat(results, axis=1, sort=False)

        # Convert the data to the requested interval
        return result.asfreq(interval, method="ffill").ffill()


class WetterdienstPredictionConnection(
    WetterdienstConnection[NodeWetterdienstPrediction], protocol="wetterdienst_prediction"
):
    """
    The WetterdienstPredictionConnection class is a connector to the Wetterdienst API
    for retrieving weather prediction data (MOSMIX). Data can only be read with
    :func:`~wetterdienst.WetterdienstPredictionConnection.read_series`.
    """

    def read_series(
        self,
        from_time: datetime,
        to_time: datetime,
        nodes: Nodes[NodeWetterdienstPrediction] | None = None,
        interval: TimeStep = 0,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """
        Read weather prediction data from the Wetterdienst API for the given nodes.
        The interval parameter is not used for prediction data, as predictions are always given hourly.

        :param from_time: Start time for the data retrieval
        :param to_time: End time for the data retrieval
        :param nodes: Nodes to read data from
        :param interval: - Not used for prediction data
        :return: Pandas DataFrame containing the data read from the connection
        """

        super().read_series(from_time, to_time, nodes, interval)
        nodes = self._validate_nodes(nodes)

        def _read_node(node: NodeWetterdienstPrediction) -> pd.Dataframe:
            request = DwdMosmixRequest(
                parameter=node.parameter,
                mosmix_type=node.mosmix_type,
                start_date=from_time,
                end_date=to_time,
                settings=self.settings,
            )
            return self.retrieve_stations(node, request)

        # We can't use a ThreadPoolExecutor here, as the Wetterdienst library uses asyncio.
        # As a result, we have to call the _read_node method directly, which causes type errors.
        results = []
        for node in nodes:
            results.append(_read_node(node))
        return pd.concat(results, axis=1, sort=False)



"""This module implements the node class, which is used to parametrize connections"""

from __future__ import annotations

import ast
import enum
import pathlib
import re
from collections.abc import Mapping
from datetime import timedelta
from sys import maxsize
from typing import TYPE_CHECKING

import attrs
import pandas as pd
from attrs import (
    converters,
    define,
    field,
    validators as vld,
)
from dateutil.parser import parse
from wetterdienst.metadata.parameter import Parameter
from wetterdienst.provider.dwd.mosmix.api import DwdMosmixParameter
from wetterdienst.provider.dwd.observation import (
    DwdObservationParameter,
    DwdObservationResolution,
)

from eta_utility import dict_get_any, get_logger, url_parse

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any, Callable, ClassVar, Final
    from urllib.parse import ParseResult

    from typing_extensions import Self, TypeAlias

    from eta_utility.type_hints import Path

default_schemes = {
    "modbus": "modbus.tcp",
    "emonio": "modbus.tcp",
    "opcua": "opc.tcp",
    "eneffco": "https",
    "local": "https",
    "entsoe": "https",
    "cumulocity": "https",
    "wetterdienst_observation": "https",
    "wetterdienst_prediction": "https",
    "forecast_solar": "https",
}


log = get_logger("connectors")


def _strip_str(value: str) -> str:
    """Convenience function to convert a string to its stripped version.

    :param value: String to convert.
    :return: Stripped string.
    """
    return value.strip()


def _lower_str(value: str) -> str:
    """Convenience function to convert a string to its stripped and lowercase version.

    :param value: String to convert.
    :return: Stripped and lowercase string.
    """
    return value.strip().lower()


def _dtype_converter(value: str) -> Callable | None:
    """Specify data type conversion functions (i.e. to convert modbus types to python).

    :param value: Data type string to convert to callable datatype converter.
    :return: Python datatype (callable).
    """
    _dtypes = {
        "boolean": bool,
        "bool": bool,
        "int": int,
        "uint32": int,
        "integer": int,
        "sbyte": int,
        "float": float,
        "double": float,
        "short": float,
        "string": str,
        "str": str,
        "bytes": bytes,
        "none": None,
        "list": list,
        "tuple": tuple,
        "dict": dict,
    }

    try:
        if value.startswith("list") or value.startswith("tuple") or value.startswith("dict"):
            value = value.split("[")[0]
        dtype = _dtypes[_lower_str(value)]
    except KeyError:
        log.warning(
            f"The specified data type ({value}) is currently not available in the datatype map and "
            f"will not be applied."
        )
        dtype = None

    return dtype


class NodeMeta(type):
    """Metaclass to define all Node classes as frozen attr dataclasses."""

    def __new__(cls, name: str, bases: tuple, namespace: dict[str, Any], **kwargs: Any) -> NodeMeta:
        attrs_args = kwargs.pop("attrs_args", {})
        new_cls = super().__new__(cls, name, bases, namespace, **kwargs)
        return define(frozen=True, slots=False, **attrs_args)(new_cls)


class Node(metaclass=NodeMeta):
    """The node objects represents a single variable. Valid keyword arguments depend on the protocol."""

    #: Name for the node.
    name: str = field(converter=_strip_str, eq=True)
    #: URL of the connection.
    url: str = field(eq=True, order=True)
    #: Parse result object of the URL (in case more post-processing is required).
    url_parsed: ParseResult = field(init=False, repr=False, eq=False, order=False)
    #: Protocol of the connection.
    protocol: str = field(repr=False, eq=False, order=False)
    #: Username for login to the connection (default: None).
    usr: str | None = field(default=None, kw_only=True, repr=False, eq=False, order=False)
    #: Password for login to the connection (default: None).
    pwd: str | None = field(default=None, kw_only=True, repr=False, eq=False, order=False)
    #: Interval
    interval: str | None = field(
        default=None, converter=converters.optional(float), kw_only=True, repr=False, eq=False, order=False
    )
    #: Data type of the node (for value conversion). Note that strings will be interpreted as utf-8 encoded. If you
    #: do not want this behaviour, use 'bytes'.
    dtype: Callable | None = field(
        default=None, converter=converters.optional(_dtype_converter), kw_only=True, repr=False, eq=False, order=False
    )

    _registry: ClassVar = {}

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Store subclass definitions to instantiate based on protocol."""
        protocol = kwargs.pop("protocol", None)
        if protocol:
            cls._registry[protocol] = cls

        return super().__init_subclass__(**kwargs)

    def __new__(cls, name: str, url: str, protocol: str, *args: Any, **kwargs: Any) -> Self:
        """Create node object of correct subclass corresponding to protocol."""
        try:
            subclass = cls._registry[protocol]
        except KeyError as error:
            raise ValueError(f"Specified an unsupported protocol: {protocol}.") from error

        # Return the correct subclass for the specified protocol
        return object.__new__(subclass)

    def __attrs_post_init__(self) -> None:
        """Add post-processing to the url, username and password information. Username and password specified during
        class init take precedence.
        """
        url, usr, pwd = url_parse(self.url, scheme=default_schemes[self.protocol])

        if self.usr is None or str(self.usr) == "nan":
            object.__setattr__(self, "usr", usr)
        object.__setattr__(self, "usr", str(self.usr) if self.usr is not None else None)

        if self.pwd is None or str(self.pwd) == "nan":
            object.__setattr__(self, "pwd", pwd)
        object.__setattr__(self, "pwd", str(self.pwd) if self.pwd is not None else None)

        object.__setattr__(self, "url", url.geturl())
        object.__setattr__(self, "url_parsed", url)

    def evolve(self, **kwargs: Any) -> Node:
        """Returns a new node instance
        by copying the current node and changing only specified keyword arguments.

        This allows for seamless node instantiation with only a few changes.

        :param kwargs: Keyword arguments to change.
        :return: New instance of the node.
        """
        return attrs.evolve(self, **kwargs)  # type: ignore

    def as_dict(self, filter_none: bool = False, **kwargs: Any) -> dict[str, Any]:
        """Return the attrs attribute values of node instance as a dict.

        :param filter_none: Filter none values, defaults to False
        :return: dict of attribute values
        """
        filter_func = self.__class__._filter_none(self) if filter_none else None
        return attrs.asdict(self, filter=filter_func, **kwargs)  # type: ignore

    def as_tuple(self, filter_none: bool = False, **kwargs: Any) -> tuple[Any, ...]:
        """Return the attrs attribute values of inst as a tuple.

        :param filter_none: Filter none values, defaults to False
        :return: tuple of attribute values
        """
        filter_func = self.__class__._filter_none(self) if filter_none else None
        return attrs.astuple(self, filter=filter_func, **kwargs)  # type: ignore

    @staticmethod
    def _filter_none(node: Node) -> Callable[[attrs.Attribute[Any], Any], bool]:
        """Return callable to filter none values, to be passed to attrs.asdict or attrs.astuple."""
        attributes = attrs.asdict(node)  # type: ignore
        non_values = {key: value for key, value in attributes.items() if value is None}
        return attrs.filters.exclude(*non_values.keys())

    @classmethod
    def from_dict(cls, dikt: Sequence[Mapping] | Mapping[str, Any], fail: bool = True) -> list[Self]:
        """Create nodes from a dictionary of node configurations. The configuration must specify the following
        fields for each node:

            * Code (or name), URL, Protocol (i.e. modbus or opcua or eneffco).
              The URL should be a complete network location identifier. Alternatively it is possible to specify the
              location in two fields: IP and Port. These should only contain the respective parts (as in only an IP
              address and only the port number).
              The IP-Address should always be given without scheme (https://).

        For local nodes no additional fields are required.

        For Modbus nodes the following additional fields are required:

            * ModbusRegisterType (or mb_register), ModbusSlave (or mb_slave), ModbusChannel (or mb_channel).

        For OPC UA nodes the following additional fields are required:

            * Identifier.

        For EnEffCo nodes the code field must be present.

        For EntsoE nodes the endpoint field must be present.

        :param dikt: Configuration dictionary.
        :param fail: Set this to false, if you would like to log errors instead of raising them.
        :return: List of Node objects.
        """

        nodes = []

        iter_ = [dikt] if isinstance(dikt, Mapping) else dikt
        for idx, lnode in enumerate(iter_):
            node = {k.strip().lower(): v for k, v in lnode.items()}

            try:
                protocol = cls._read_dict_protocol(node)
            except KeyError as e:
                text = f"Error reading node protocol in row {idx + 1}: {e}."
                if fail:
                    raise KeyError(text) from e
                log.error(text)
                continue

            try:
                node_class = cls._registry[protocol.strip().lower()]
            except KeyError as e:
                text = f"Specified an unsupported protocol in row {idx + 1}: {protocol}."
                if fail:
                    raise ValueError(text) from e
                log.error(text)
                continue

            try:
                nodes.append(node_class._from_dict(node))
            except (TypeError, KeyError) as e:
                text = f"Error while reading the configuration data for node in row {idx + 1}: {e}."
                if fail:
                    raise TypeError(text) from e
                log.error(text)

        return nodes

    @staticmethod
    def _read_dict_info(node: dict[str, Any]) -> tuple[str, str, str, str, int]:
        """Read general info about a node from a dictionary.

        :param node: dictionary containing node information.
        :return: name, pwd, url, usr of the node
        """
        # Read name first
        try:
            name = str(dict_get_any(node, "code", "name"))
            if name == "nan" or name is None:
                raise KeyError
        except KeyError as e:
            raise KeyError("Name or Code must be specified for all nodes in the dictionary.") from e
        # Find URL or IP and port
        if "url" in node and node["url"] is not None and str(node["url"]) not in {"nan", ""}:
            url = node["url"].strip()
        elif "ip" in node and node["ip"] is not None and str(node["ip"]) not in {"nan", ""}:
            _port = dict_get_any(node, "port", fail=False, default="")
            port = "" if _port in {None, ""} or str(_port) == "nan" else f":{int(_port)}"
            url = f"{dict_get_any(node, 'ip')}{port}"
        else:
            url = None
        usr = dict_get_any(node, "username", "user", "usr", fail=False)
        pwd = dict_get_any(node, "password", "pwd", "pw", fail=False)
        interval = dict_get_any(node, "interval", fail=False)
        return name, pwd, url, usr, interval

    @staticmethod
    def _read_dict_protocol(node: dict[str, Any]) -> str:
        try:
            protocol = str(dict_get_any(node, "protocol"))
            if protocol == "nan" or protocol is None:
                raise KeyError
        except KeyError as e:
            raise KeyError("Protocol must be specified for all nodes in the dictionary.") from e

        return protocol

    @staticmethod
    def _try_dict_get_any(dikt: dict[str, Any], *names: str) -> Any:
        """Get any of the specified items from the node, if any are available. The function will return
        the first value it finds, even if there are multiple matches.

        This function will output sensible error messages, when the values are not found.

        :param dikt: Dictionary of the node to get values from.
        :param names: Item names to look for.
        :return: Value from dictionary.
        """
        try:
            value = dict_get_any(dikt, *names, fail=True)
        except KeyError as e:
            log.error(f"For the node, the field '{names[0]}' must be specified or check the correct spelling.")
            raise KeyError(
                "The required parameter for the node configuration was not found (see log). "
                "Could not load config "
                "file. "
            ) from e

        return value

    @classmethod
    def from_excel(cls, path: Path, sheet_name: str, fail: bool = True) -> list[Self]:
        """
        Method to read out nodes from an Excel document. The document must specify the following fields:

            * Code, IP, Port, Protocol (modbus or opcua or eneffco).

        For Modbus nodes the following additional fields are required:

            * ModbusRegisterType, ModbusByte, ModbusChannel.

        For OPC UA nodes the following additional fields are required:

            * Identifier.

        For EnEffCo nodes the Code field must be present.

        The IP-Address should always be given without scheme (https://).

        :param path: Path to Excel document.
        :param sheet_name: name of Excel sheet, which will be read out.
        :param fail: Set this to false, if you would like to log errors instead of raising them.
        :return: List of Node objects.
        """

        file = path if isinstance(path, pathlib.Path) else pathlib.Path(path)
        input_ = pd.read_excel(file, sheet_name=sheet_name, dtype=str)

        return cls.from_dict(list(input_.to_dict("index").values()), fail)

    @classmethod
    def get_eneffco_nodes_from_codes(cls, code_list: Sequence[str], eneffco_url: str) -> list[Self]:
        """
        Utility function to retrieve Node objects from a list of EnEffCo Codes (Identifiers).

        .. deprecated:: v2.0.0
            Use the *from_ids* function of the EnEffCoConnection Class instead.

        :param code_list: List of EnEffCo identifiers to create nodes from.
        :param eneffco_url: URL to the EnEffCo system.
        :return: List of EnEffCo nodes.
        """
        nodes = []
        for code in code_list:
            nodes.append(cls(name=code, url=eneffco_url, protocol="eneffco", eneffco_code=code))
        return nodes


class NodeLocal(Node, protocol="local"):
    """Local Node (no specific protocol), useful for example to manually provide data to subscription handlers."""

    def __attrs_post_init__(self) -> None:
        """Ensure username and password are processed correctly."""
        super().__attrs_post_init__()

    @classmethod
    def _from_dict(cls, dikt: dict[str, Any]) -> NodeLocal:
        """Create a local node from a dictionary of node information.

        :param dikt: dictionary with node information.
        :return: NodeLocal object.
        """
        name, pwd, url, usr, interval = cls._read_dict_info(dikt)
        try:
            return cls(name, url, "local", usr=usr, pwd=pwd, interval=interval)
        except (TypeError, AttributeError) as e:
            raise TypeError(f"Could not convert all types for node {name}") from e


def _mb_endianness_converter(value: str) -> str:
    """Convert some values for mb_byteorder.

    :param value: Value to be converted to mb_byteorder
    :return: mb_byteorder corresponding to correct scheme.
    """
    value = _lower_str(value)
    if value in {"little", "littleendian"}:
        return "little"

    if value in {"big", "bigendian"}:
        return "big"

    return ""


class NodeModbus(Node, protocol="modbus"):
    """Node for the Modbus protocol."""

    #: Modbus Slave ID
    mb_slave: int | None = field(kw_only=True, default=32, converter=int)
    #: Modbus Register name. One of input, discrete_input, coils and holding. Note that only coils and
    #: holding can be written to.
    mb_register: str = field(
        kw_only=True, converter=_lower_str, validator=vld.in_(("input", "discrete_input", "coils", "holding"))
    )
    #: Modbus Channel (Address of the value)
    mb_channel: int = field(kw_only=True, converter=int)
    #: Length of the value in bits (default 32). This determines, how much data is read from the server. The
    #: value must be a multiple of 16.
    mb_bit_length: int = field(kw_only=True, default=32, converter=int, validator=vld.ge(1))

    #: Byteorder of values returned by modbus
    mb_byteorder: str = field(kw_only=True, converter=_mb_endianness_converter, validator=vld.in_(("little", "big")))
    #: Wordorder of values returned by modbus
    mb_wordorder: str = field(
        default="big", kw_only=True, converter=_mb_endianness_converter, validator=vld.in_(("little", "big"))
    )

    def __attrs_post_init__(self) -> None:
        """Add default port to the URL and convert mb_byteorder values."""
        super().__attrs_post_init__()

        # Set port to default 502 if it was not explicitly specified
        if not isinstance(self.url_parsed.port, int):
            url = self.url_parsed._replace(netloc=f"{self.url_parsed.hostname}:502")
            object.__setattr__(self, "url", url.geturl())
            object.__setattr__(self, "url_parsed", url)

    @classmethod
    def _from_dict(cls, dikt: dict[str, Any]) -> Self:
        """Create a modbus node from a dictionary of node information.

        :param dikt: dictionary with node information.
        :return: NodeModbus object.
        """
        name, pwd, url, usr, interval = cls._read_dict_info(dikt)
        # Initialize node if protocol is 'modbus'
        try:
            mb_register = cls._try_dict_get_any(dikt, "mb_register", "modbusregistertype")
            mb_channel = cls._try_dict_get_any(dikt, "mb_channel", "modbuschannel")
            mb_byteorder = cls._try_dict_get_any(dikt, "mb_byteorder", "modbusbyteorder")
            mb_wordorder = dict_get_any(dikt, "mb_wordorder", "modbuswordorder", fail=False, default="big")
            mb_slave = dict_get_any(dikt, "mb_slave", "modbusslave", fail=False, default=32)
            mb_bit_length = dict_get_any(dikt, "mb_bit_length", "mb_bitlength", fail=False, default=32)
            dtype = dict_get_any(dikt, "dtype", "datentyp", fail=False)
        except KeyError as e:
            raise KeyError(
                f"The required parameter for the node configuration was not found (see log). The node {name} could "
                f"not load."
            ) from e
        try:
            return cls(
                name,
                url,
                "modbus",
                usr=usr,
                pwd=pwd,
                mb_register=mb_register,
                mb_slave=mb_slave,
                mb_channel=mb_channel,
                mb_bit_length=mb_bit_length,
                mb_byteorder=mb_byteorder,
                mb_wordorder=mb_wordorder,
                dtype=dtype,
                interval=interval,
            )
        except (TypeError, AttributeError) as e:
            raise TypeError(f"Could not convert all types for node {name}.") from e


class NodeOpcUa(Node, protocol="opcua"):
    """Node for the OPC UA protocol."""

    #: Node ID of the OPC UA Node.
    opc_id: str | None = field(default=None, kw_only=True, converter=converters.optional(_strip_str))
    #: Path to the OPC UA node.
    opc_path_str: str | None = field(
        default=None, kw_only=True, converter=converters.optional(_strip_str), repr=False, eq=False, order=False
    )
    #: Namespace of the OPC UA Node.
    opc_ns: int | None = field(default=None, kw_only=True, converter=converters.optional(_lower_str))

    # Additional fields which will be determined automatically
    #: Type of the OPC UA Node ID Specification.
    opc_id_type: str = field(
        init=False, converter=str, validator=vld.in_(("i", "s")), repr=False, eq=False, order=False
    )
    #: Name of the OPC UA Node.
    opc_name: str = field(init=False, repr=False, eq=False, order=False, converter=str)
    #: Path to the OPC UA node in list representation. Nodes in this list can be used to access any
    #: parent objects.
    opc_path: list[NodeOpcUa] = field(init=False, repr=False, eq=False, order=False)

    def __attrs_post_init__(self) -> None:
        """Add default port to the URL and convert mb_byteorder values."""
        super().__attrs_post_init__()

        # Set port to default 4840 if it was not explicitly specified
        if not isinstance(self.url_parsed.port, int):
            url = self.url_parsed._replace(netloc=f"{self.url_parsed.hostname}:4840")
            object.__setattr__(self, "url", url.geturl())
            object.__setattr__(self, "url_parsed", url)

        # Determine, which values to use for initialization and set values
        if self.opc_id is not None:
            try:
                parts = self.opc_id.split(";")
            except ValueError as e:
                raise ValueError(
                    f"When specifying opc_id, make sure it follows the format ns=2;s=.path (got {self.opc_id})."
                ) from e
            for part in parts:
                try:
                    key, val = part.split("=")
                except ValueError as e:
                    raise ValueError(
                        f"When specifying opc_id, make sure it follows the format ns=2;s=.path (got {self.opc_id})."
                    ) from e

                if key.strip().lower() == "ns":
                    object.__setattr__(self, "opc_ns", int(val))
                else:
                    object.__setattr__(self, "opc_id_type", key.strip().lower())
                    object.__setattr__(self, "opc_path_str", val.strip())

            object.__setattr__(self, "opc_id", f"ns={self.opc_ns};{self.opc_id_type}={self.opc_path_str}")

        elif self.opc_path_str is not None and self.opc_ns is not None:
            object.__setattr__(self, "opc_id_type", "s")
            object.__setattr__(self, "opc_id", f"ns={self.opc_ns};s={self.opc_path_str}")
        else:
            raise ValueError("Specify opc_id or opc_path_str and ns for OPC UA nodes.")

        # Determine the name and path of the opc node
        split_path = (
            self.opc_path_str.rsplit(".", maxsplit=len(self.opc_path_str.split(".")) - 2)  # type: ignore
            if self.opc_path_str[0] == "."  # type: ignore
            else self.opc_path_str.split(".")  # type: ignore
        )

        object.__setattr__(self, "opc_name", split_path[-1].split(".")[-1])
        path = []
        for k in range(len(split_path) - 1):
            path.append(
                Node(
                    split_path[k].strip(" ."),
                    self.url,
                    "opcua",
                    usr=self.usr,
                    pwd=self.pwd,
                    opc_id="ns={};s={}".format(self.opc_ns, ".".join(split_path[: k + 1])),
                )
            )
        object.__setattr__(self, "opc_path", path)

    @classmethod
    def _from_dict(cls, dikt: dict[str, Any]) -> Self:
        """Create an opcua node from a dictionary of node information.

        :param dikt: dictionary with node information.
        :return: NodeOpcUa object.
        """
        name, pwd, url, usr, interval = cls._read_dict_info(dikt)

        opc_id = dict_get_any(dikt, "opc_id", "identifier", "identifier", fail=False)
        dtype = dict_get_any(dikt, "dtype", "datentyp", fail=False)

        if opc_id is None:
            opc_ns = dict_get_any(dikt, "opc_ns", "namespace", "ns", fail=False)
            opc_path_str = dict_get_any(dikt, "opc_path", "path", fail=False)
            try:
                return cls(
                    name,
                    url,
                    "opcua",
                    usr=usr,
                    pwd=pwd,
                    opc_ns=opc_ns,
                    opc_path_str=opc_path_str,
                    dtype=dtype,
                    interval=interval,
                )
            except (TypeError, AttributeError) as e:
                raise TypeError(
                    f"Could not convert all types for node {name}. Either the 'node_id' or the 'opc_ns' "
                    f"and 'opc_path' must be specified."
                ) from e
        else:
            try:
                return cls(name, url, "opcua", usr=usr, pwd=pwd, opc_id=opc_id, dtype=dtype, interval=interval)
            except (TypeError, AttributeError) as e:
                raise TypeError(
                    f"Could not convert all types for node {name}. Either the 'node_id' or the 'opc_ns' "
                    f"and 'opc_path' must be specified."
                ) from e

    def evolve(self, **kwargs: Any) -> Node:
        """Returns a new node instance
        by copying the current node and changing only specified keyword arguments.

        This allows for seamless node instantiation with only a few changes.

        Adjusted attributes handling according to OpcUa node instantiation logic as in '__attrs_post_init__'.

        :param kwargs: Keyword arguments to change.
        :return: New instance of the node.
        """
        # Ensure that opc_id is not set if opc_ns and opc_path_str are set, to avoid postprocessing conflicts
        if kwargs.get("opc_id") is not None or self.opc_id is not None:
            kwargs["opc_ns"] = None
            kwargs["opc_path_str"] = None
        else:
            kwargs["opc_ns"] = str(kwargs.get("opc_ns", self.opc_ns))

        return super().evolve(**kwargs)


class NodeEnEffCo(Node, protocol="eneffco"):
    """Node for the EnEffCo API."""

    #: EnEffCo datapoint code / ID.
    eneffco_code: str = field(kw_only=True, converter=str)

    def __attrs_post_init__(self) -> None:
        """Ensure username and password are processed correctly."""
        super().__attrs_post_init__()

    @classmethod
    def _from_dict(cls, dikt: dict[str, Any]) -> Self:
        """Create a EnEffCo node from a dictionary of node information.

        :param dikt: dictionary with node information.
        :return: NodeEnEffCo object.
        """
        name, pwd, url, usr, interval = cls._read_dict_info(dikt)
        try:
            code = cls._try_dict_get_any(dikt, "code", "eneffco_code")
        except KeyError as e:
            raise KeyError(
                f"The required parameter for the node configuration was not found (see log). The node {name} could "
                f"not load."
            ) from e

        try:
            return cls(name, url, "eneffco", usr=usr, pwd=pwd, eneffco_code=code, interval=interval)
        except (TypeError, AttributeError) as e:
            raise TypeError(f"Could not convert all types for node {name}.") from e


class NodeEntsoE(Node, protocol="entsoe"):
    """Node for the EntsoE API (see `ENTSO-E Transparency Platform API
    <https://transparency.entsoe.eu/content/static_content/Static%20content/web%20api/Guide.html>`_).

    .. list-table:: **Available endpoint**
        :widths: 25 35
        :header-rows: 1

        * - Endpoint
          - Description
        * - ActualGenerationPerType
          - Actual Generation Per Energy Type
        * - Price
          - Price day ahead

    Currently, there is only two endpoints available, due to the parameter managing required by the API documentation.
    The other possible endpoints are listed in

    `eta_utility.connectors.entso_e._ConnectionConfiguration._doc_types`

    .. list-table:: **Main bidding zone**
        :widths: 15 25
        :header-rows: 1

        * - Bidding Zone
          - Description
        * - DEU-LUX
          - Deutschland-Luxemburg

    The other possible bidding zones are listed in

    `eta_utility.connectors.entso_e._ConnectionConfiguration._bidding_zones`

    """

    #: REST endpoint.
    endpoint: str = field(kw_only=True, converter=str)
    #: Bidding zone.
    bidding_zone: str = field(kw_only=True, converter=str)

    def __attrs_post_init__(self) -> None:
        """Ensure username and password are processed correctly."""
        super().__attrs_post_init__()

    @classmethod
    def _from_dict(cls, dikt: dict[str, Any]) -> NodeEntsoE:
        """Create an EntsoE node from a dictionary of node information.

        :param dikt: dictionary with node information.
        :return: NodeEntsoE object.
        """
        name, pwd, url, usr, interval = cls._read_dict_info(dikt)

        try:
            endpoint = cls._try_dict_get_any(dikt, "endpoint")
            bidding_zone = cls._try_dict_get_any(dikt, "bidding zone", "bidding_zone", "zone")
        except KeyError as e:
            raise KeyError(
                f"The required parameter for the node configuration was not found (see log). The node {name} could "
                f"not load."
            ) from e

        try:
            return cls(
                name, url, "entsoe", usr=usr, pwd=pwd, endpoint=endpoint, bidding_zone=bidding_zone, interval=interval
            )
        except (TypeError, AttributeError) as e:
            raise TypeError(f"Could not convert all types for node {name}.") from e


class NodeCumulocity(Node, protocol="cumulocity"):
    """Node for the Cumulocity API."""

    # parameters for reading/writing from/to cumulocity nodes
    device_id: str = field(kw_only=True, converter=str)
    measurement: str = field(kw_only=True, converter=str)
    fragment: str = field(kw_only=True, converter=str, default="")

    def __attrs_post_init__(self) -> None:
        """Ensure username and password are processed correctly."""
        super().__attrs_post_init__()

    @classmethod
    def _from_dict(cls, dikt: dict[str, Any]) -> NodeCumulocity:
        """Create a Cumulocity node from a dictionary of node information.

        :param dikt: dictionary with node information.
        :return: NodeCumulocity object.
        """
        name, pwd, url, usr, interval = cls._read_dict_info(dikt)
        try:
            device_id = cls._try_dict_get_any(dikt, "id", "device_id")
        except KeyError as e:
            raise KeyError(
                f"The required parameter for the node configuration was not found (see log). The node {name} could "
                f"not load."
            ) from e
        try:
            measurement = cls._try_dict_get_any(dikt, "measurement", "Measurement")
        except KeyError as e:
            raise KeyError(
                f"The required parameter for the node configuration was not found (see log). The node {name} could "
                f"not load."
            ) from e

        try:
            fragment = cls._try_dict_get_any(dikt, "fragment", "Fragment")
        except KeyError:
            fragment = ""
        try:
            return cls(
                name,
                url,
                "cumulocity",
                usr=usr,
                pwd=pwd,
                device_id=device_id,
                measurement=measurement,
                fragment=fragment,
            )
        except (TypeError, AttributeError) as e:
            raise TypeError(f"Could not convert all types for node {name}.") from e


class NodeWetterdienst(Node):
    """
    Basic Node for the Wetterdienst API.
    This class is not meant to be used directly, but to be subclassed by
    NodeWetterdienstObservation and NodeWetterdienstPrediction.
    """

    #: Parameter to read from wetterdienst (e.g HUMIDITY or TEMPERATURE_AIR_200)
    parameter: str = field(kw_only=True, converter=str.upper)

    #: The id of the weather station
    station_id: str | None = field(default=None, kw_only=True)
    #: latitude and longitude (not necessarily a weather station)
    latlon: str | None = field(default=None, kw_only=True)
    #: Number of stations to be used for the query
    number_of_stations: int | None = field(default=None, kw_only=True)

    def __attrs_post_init__(self) -> None:
        """Ensure that all required parameters are present."""
        # Set same default URL for all Wetterdienst nodes
        object.__setattr__(self, "url", "https://opendata.dwd.de")
        super().__attrs_post_init__()
        if self.station_id is None and (self.latlon is None or self.number_of_stations is None):
            raise ValueError(
                "The required parameter 'station_id' or 'latlon' and 'number_of_stations' for the node configuration "
                "was not found. The node could not load."
            )
        parameters = [item.name for item in Parameter]
        if self.parameter not in parameters:
            raise ValueError(
                f"Parameter {self.parameter} is not valid. Valid parameters can be found here:"
                f"https://wetterdienst.readthedocs.io/en/latest/data/parameters.html"
            )

    @classmethod
    def _get_params(cls, dikt: dict[str, Any]) -> dict[str, Any]:
        """Get the common parameters for a Wetterdienst node.

        :param dikt: dictionary with node information.
        :return: dict with: parameter, station_id, latlon, number_of_stations
        """
        return {
            "parameter": dikt.get("parameter"),
            "station_id": dikt.get("station_id"),
            "latlon": dikt.get("latlon"),
            "number_of_stations": dikt.get("number_of_stations"),
        }


class NodeWetterdienstObservation(NodeWetterdienst, protocol="wetterdienst_observation"):
    """
    Node for the Wetterdienst API to get weather observations.
    For more information see: https://wetterdienst.readthedocs.io/en/latest/data/coverage/dwd/observation.html
    """

    #: Redeclare interval attribute, but don't allow it to be optional
    interval: str = field(converter=converters.optional(float), kw_only=True, repr=False, eq=False, order=False)

    def __attrs_post_init__(self) -> None:
        super().__attrs_post_init__()
        resolution = self.convert_interval_to_resolution(self.interval)
        # Sort out the parameters by resolution
        available_params = DwdObservationParameter[resolution]
        available_params = [param.name for param in available_params if type(param) is not enum.EnumMeta]

        # If the parameter is not in the available parameters for the resolution, generate a list
        # of available resolutions for the parameter and raise an error
        if self.parameter not in available_params:
            available_resolutions = []
            for resolution in DwdObservationResolution:
                params = DwdObservationParameter[resolution.name]  # type: ignore
                if self.parameter in [param.name for param in params if type(param) is not enum.EnumMeta]:
                    available_resolutions.append(resolution.name)  # type: ignore
            if len(available_resolutions) == 0:
                raise ValueError(f"Parameter {self.parameter} is not a valid observation parameter.")
            raise ValueError(
                f"Parameter {self.parameter} is not valid for the given resolution. "
                f"Valid resolutions for parameter {self.parameter} are: "
                f"{available_resolutions}"
            )

    @classmethod
    def _from_dict(cls, dikt: dict[str, Any]) -> NodeWetterdienstObservation:
        """Create a WetterdienstObservation node from a dictionary of node information.

        :param dikt: dictionary with node information.
        :return: NodeWetterdienst object.
        """
        name, _, _, _, interval = cls._read_dict_info(dikt)
        params = cls._get_params(dikt)
        try:
            return cls(name, "", "wetterdienst_observation", interval=interval, **params)
        except (TypeError, AttributeError) as e:
            raise TypeError(f"Could not convert all types for node {name}.") from e

    @staticmethod
    def convert_interval_to_resolution(interval: int | str | timedelta) -> str:
        resolutions = {
            60: "MINUTE_1",
            300: "MINUTE_5",
            600: "MINUTE_10",
            3600: "HOURLY",
            28800: "SUBDAILY",  # not 8h intervals, measured at 7am, 2pm, 9pm
            86400: "DAILY",
            2592000: "MONTHLY",
            31536000: "ANNUAL",
        }
        interval = int(interval.total_seconds()) if isinstance(interval, timedelta) else int(interval)
        if interval not in resolutions:
            raise ValueError(f"Interval {interval} not supported. Must be one of {list(resolutions.keys())}")
        return resolutions[interval]


class NodeWetterdienstPrediction(NodeWetterdienst, protocol="wetterdienst_prediction"):
    """
    Node for the Wetterdienst API to get weather predictions.
    For more information see: https://wetterdienst.readthedocs.io/en/latest/data/coverage/dwd/mosmix.html
    """

    #: Type of the MOSMIX prediction. Either 'SMALL' or 'LARGE'
    mosmix_type: str = field(kw_only=True, converter=str.upper, validator=vld.in_(("SMALL", "LARGE")))

    def __attrs_post_init__(self) -> None:
        super().__attrs_post_init__()
        # Sort out the parameters by resolution
        params = DwdMosmixParameter[self.mosmix_type]
        # Create list of available parameters, enums are excluded because they are datasets
        available_params = [param.name for param in params if type(param) is not enum.EnumMeta]

        if self.parameter not in available_params:
            raise ValueError(
                f"Parameter {self.parameter} is not valid for the given resolution."
                f"Valid parameters for resolution {self.mosmix_type} can be found here:"
                f"https://wetterdienst.readthedocs.io/en/latest/data/coverage/dwd/mosmix/hourly.html"
            )

    @classmethod
    def _from_dict(cls, dikt: dict[str, Any]) -> NodeWetterdienstPrediction:
        """Create a WetterdienstPrediction node from a dictionary of node information.

        :param dikt: dictionary with node information.
        :return: NodeWetterdienst object.
        """
        name, _, _, _, _ = cls._read_dict_info(dikt)
        params = cls._get_params(dikt)
        mosmix_type = dikt.get("mosmix_type")
        try:
            return cls(name, "", "wetterdienst_prediction", mosmix_type=mosmix_type, **params)
        except (TypeError, AttributeError) as e:
            raise TypeError(f"Could not convert all types for node {name}.") from e


class EmonioConstants:
    """Dict constants for the Emonio API."""

    #: Mapping of parameters to addresses
    PARAMETER_MAP: Final[dict[int, list[str]]] = {
        0: ["VRMS", "V_RMS", "Voltage", "V", "Spannung"],
        2: ["IRMS", "I_RMS", "Current", "I", "Strom"],
        4: ["WATT", "Power", "W", "Leistung", "Wirkleistung"],
        6: ["VAR", "Reactive Power", "VAR", "Blindleistung"],
        8: ["VA", "Apparent Power", "VA", "Scheinleistung"],
        10: ["FREQ", "Frequency", "Hz", "Frequenz"],
        12: ["KWH", "Energy", "kWh", "Energie"],
        14: ["PF", "Power Factor", "PF", "Leistungsfaktor"],
        20: ["VRMS MIN", "VRMS_MIN", "Voltage Min", "V Min", "Spannung Min"],
        22: ["VRMS MAX", "VRMS_MAX", "Voltage Max", "V Max", "Spannung Max"],
        24: ["IRMS MIN", "IRMS_MIN", "Current Min", "I Min", "Strom Min"],
        26: ["IRMS MAX", "IRMS_MAX", "Current Max", "I Max", "Strom Max"],
        28: ["WATT MIN", "WATT_MIN", "Power Min", "W Min", "Leistung Min"],
        30: ["WATT MAX", "WATT_MAX", "Power Max", "W Max", "Leistung Max"],
        500: ["Temp", "degree", "Temperature", "°C", "Temperatur"],
        800: ["Impulse", "Impuls"],
    }
    #: Create dictionary with all upper cased parameters
    UPPER_CASED: Final[dict[int, list[str]]] = {
        adr: [par.upper() for par in par_list] for (adr, par_list) in PARAMETER_MAP.items()
    }
    #: Mapping of phases to address offsets
    PHASE_MAP: Final[dict[str, int]] = {
        "a": 0,
        "b": 100,
        "c": 200,
        "abc": 300,
    }


class NodeEmonio(Node, protocol="emonio"):
    """
    Node for the emonio. The parameter to read is specified by the name of the node.
    Available parameters are defined in the parameter_map class attribute.
    Additionally, the phase of the parameter can be specified, with 'a', 'b', 'c' or 'abc'.

    https://wiki.emonio.de/de/Emonio_P3
    """

    #: Modbus address of the parameter to read
    address: int = field(default=-1, kw_only=True, converter=int)
    #: Phase of the parameter (a, b, c). If not set, all phases are read
    phase: str = field(default="abc", kw_only=True, converter=_lower_str, validator=vld.in_(("a", "b", "c", "abc")))

    def __attrs_post_init__(self) -> None:
        """Ensure that all required parameters are present and valid."""
        super().__attrs_post_init__()

        if self.address == -1:
            address = self._translate_name()
            object.__setattr__(self, "address", address)
        assert self.address != -1

        _parameter = self.address % 100
        _phase = self.address // 100 * 100
        # Validate address
        if self.address in {500, 800}:
            pass
        elif _parameter not in EmonioConstants.PARAMETER_MAP or _phase not in EmonioConstants.PHASE_MAP.values():
            raise ValueError(f"Address {self.address} for node {self.name} is not valid.")
        elif _parameter >= 20 and _parameter <= 30 and _phase == 300:
            raise ValueError("Phase must be set for MIN/MAX values")

    def _translate_name(self) -> int:
        """Translate the name of the node to the correct parameter name.

        :return: Modbus address of the parameter.
        """
        parameter: int | None = None
        phase: int | None = None
        # Try to find matching parameter for the name
        for address in EmonioConstants.UPPER_CASED:
            # e.g. Server1.Voltage -> VOLTAGE
            parameter_str = self.name.split(".")[-1].upper()
            if parameter_str in EmonioConstants.UPPER_CASED[address]:
                parameter = address
                log.debug(f"Parameter {parameter_str} found at address {address}")
                break
        # If no parameter was found, raise an error
        if parameter is None:
            raise ValueError(f"Parameter for node {self.name} not found, name is not valid.")

        # Temperature and Impulse values do not have a phase
        if parameter in (500, 800):
            return parameter

        # Phase is set to 0, 100, 200 or 300. (300 is default)
        phase = EmonioConstants.PHASE_MAP[self.phase]

        # Return correct address (by adding the phase offset to the parameter)
        return parameter + phase

    @classmethod
    def _from_dict(cls, dikt: dict[str, Any]) -> NodeEmonio:
        """Create an Emonio node from a dictionary of node information.

        :param dikt: dictionary with node information.
        :return: NodeEmonio object.
        """
        name, _, url, _, interval = cls._read_dict_info(dikt)

        phase = dikt.get("phase", "abc")
        phase = "abc" if pd.isna(phase) else phase

        address = dikt.get("address")
        address = -1 if pd.isna(address) else address
        try:
            return cls(name, url, "emonio", interval=interval, phase=phase, address=address)
        except (TypeError, AttributeError) as e:
            raise TypeError(f"Could not convert all types for node {name}.") from e


# Forecast.Solar API Node


def _convert_list(_type: TypeAlias) -> Callable:
    """Convert an optional list of values to a single value or a list of values.

    :param _type: type to convert the values to.
    :return: converter function.
    """

    def converter(value: Any) -> _type | list[_type]:
        try:
            if isinstance(value, str):
                value = ast.literal_eval(value)
            if isinstance(value, list):
                return [_type(val) for val in value]
            return _type(value)
        except ValueError as e:
            raise ValueError(f"Could not convert value to {_type} ({value}).") from e

    return converter


def _check_api_key(instance, attribute, value):  # type: ignore[no-untyped-def]
    """attrs validator to check if the API key is set."""
    if re.match(r"[A-Za-z0-9]{16}", value) is None:
        raise ValueError("API key must be a 16 character long alphanumeric string.")


def _check_plane(_type: TypeAlias, lower: int, upper: int) -> Callable:
    """Return an attrs validator for plane related attributes (declination, azimuth, kwp).
    Checks if the value is between the lower and upper bounds.

    :param _type: type of the parameter
    :param lower: lower bound
    :param upper: upper bound
    """

    def validator(instance, attribute, value):  # type: ignore[no-untyped-def]
        for val in [value] if not isinstance(value, list) else value:
            if not isinstance(val, _type):
                raise ValueError(f"Parameter must be of type {_type} ({(val, type(val))}).")
            if not (lower <= val <= upper):
                raise ValueError(f"Parameter must be between {lower} and {upper} ({val}).")

    return validator


def _check_horizon(instance, attribute, value):  # type: ignore[no-untyped-def]
    """attrs validator to check if horizon attribute corresponds to the API requirements."""
    if not isinstance(value, list) or value != 0:
        raise ValueError("Horizon must be a list, 0 (To suppress horizon usage *at all*) or None")
    if len(value) < 4:
        raise ValueError("Horizon must contain at least 4 values.")
    if not isinstance((360 / len(value)), int):
        raise ValueError("Make sure that 360 / <your count of values> is an integer.")
    if not all(isinstance(i, (int, float)) and 0 <= i <= 90 for i in value):
        raise ValueError("Each value in horizon must be a number between 0 and 90.")


def _check_php_datetime(instance, attribute, value):  # type: ignore[no-untyped-def]
    """attrs validator to check if the value is a valid PHP DateTime format."""
    try:
        parse(value)
    except ValueError as e:
        raise ValueError("'start' value must be a valid PHP DateTime format.") from e


def _forecast_solar_transform(cls: attrs.AttrsInstance, fields: list[attrs.Attribute]) -> list[attrs.Attribute]:
    """Transform the fields of the ForecastSolar node class.

    :param fields: list of fields to transform.
    :return: transformed fields.
    """
    for _field in fields:
        # Skip fields from superclass and reset original kw_only value
        if _field.name in ["usr", "pwd", "interval", "dtype", "name", "url", "url_parsed", "protocol"]:
            if _field.name in ["name", "url", "url_parsed", "protocol"]:
                object.__setattr__(_field, "kw_only", False)
            continue

        # Unpack field's type hints and convert them appropriately
        types = tuple(map(_dtype_converter, _field.type.split(" | ")))  # type: ignore[union-attr]

        # Create and append type validator to existing validators
        _vlds = [vld.instance_of(types)]  # type: ignore
        if _field.validator:
            _vlds.extend(
                [*_field.validator._validators]  # type: ignore
                if isinstance(_field.validator, type(vld.and_()))
                else [_field.validator]
            )
        all_validators = vld.and_(*_vlds)

        # If None in type hints, make the field optional
        if None in types:
            all_validators = vld.optional(all_validators)
            if _field.converter:
                object.__setattr__(_field, "converter", converters.optional(_field.converter))
            object.__setattr__(_field, "default", _field.default or None)
        object.__setattr__(_field, "validator", all_validators)

    return fields


attrs_args = {"kw_only": True, "field_transformer": _forecast_solar_transform}


class NodeForecastSolar(Node, protocol="forecast_solar", attrs_args=attrs_args):
    """
    Node for using the Forecast.Solar API.

    Mandatory parameters are:

    * The location of the forecast solar plane(s): **latitude**, **longitude**,
    * Plane parameters: **declination**, **azimuth** and **kwp**.

    Additionally **api_key** must be set for endpoints other than 'estimate',
    multiple planes or if requests capacity is exceeded.

    For multiple planes, the parameters shall be passed as lists of the same length
    (e.g. [0, 30], [180, 180], [5, 5]).

    """

    # URL PARAMETERS
    # ----------------

    #: API key for the Forecast.Solar API; string
    api_key: str | None = field(repr=False, converter=str, validator=_check_api_key, metadata={"QUERY_PARAM": False})
    #: Endpoint in (estimate, history, clearsky), defaults to estimate; string
    endpoint: str = field(
        default="estimate",
        converter=str,
        validator=vld.in_(("estimate", "check", "history", "clearsky")),
        metadata={"QUERY_PARAM": False},
    )
    #: What data to query, i.e. only watts, watt hours, watt hours per period or watt hours per day; string
    data: str | None = field(
        default="watts",
        converter=str,
        validator=vld.in_(("watts", "watthours", "watthoursperperiod", "watthoursperday")),
        metadata={"QUERY_PARAM": False},
    )
    #: Latitude of plane location, -90 (south) … 90 (north); handled with a precision of 0.0001 or abt. 10 m
    latitude: int = field(converter=int, validator=[vld.ge(-90), vld.le(90)], metadata={"QUERY_PARAM": False})
    #: Longitude of plane location, -180 (west) … 180 (east); handled with a precision of 0.0001 or abt. 10 m
    longitude: int = field(converter=int, validator=[vld.ge(-180), vld.le(180)], metadata={"QUERY_PARAM": False})
    #: Plane declination, 0 (horizontal) … 90 (vertical) - always in relation to earth's surface; integer
    declination: int | list[int] = field(
        converter=_convert_list(int),
        validator=_check_plane(int, 0, 90),
        metadata={"QUERY_PARAM": False},
        eq=False,  # Exclude from __hash__
    )
    #: Plane azimuth, -180 … 180 (-180 = north, -90 = east, 0 = south, 90 = west, 180 = north); integer
    azimuth: int | list[int] = field(
        converter=_convert_list(int),
        validator=_check_plane(int, -180, 180),
        metadata={"QUERY_PARAM": False},
        eq=False,  # Exclude from __hash__
    )
    #: Installed modules power of plane in kilo watt; float
    kwp: float | list[float] = field(
        converter=_convert_list(float),
        validator=_check_plane(float, 0, maxsize),
        metadata={"QUERY_PARAM": False},
        eq=False,  # Exclude from __hash__
    )

    # QUERY PARAMETERS
    # ----------------
    #: Format of timestamps in the response, see API doc for values; string
    time: str = field(
        default="utc",
        converter=str,
        validator=vld.in_(("utc", "iso8601", "rfc2822", "seconds", "milliseconds", "nanoseconds")),
        metadata={"QUERY_PARAM": True},
    )
    #: Forecast for full day or only sunrise to sunset, 0|1; int
    no_sun: int | None = field(default=None, validator=vld.in_((0, 1)), metadata={"QUERY_PARAM": True})
    #: Damping factor for the morning forecast solar API
    damping_morning: float | None = field(
        default=None, converter=float, validator=[vld.ge(0.0), vld.le(1.0)], metadata={"QUERY_PARAM": True}
    )
    #: Damping factor for the evening forecast solar API
    damping_evening: float | None = field(
        default=None, converter=float, validator=[vld.ge(0.0), vld.le(1.0)], metadata={"QUERY_PARAM": True}
    )
    #: Horizon information; string, (comma-separated list of numerics) See API doc
    horizon: int | list[int] | None = field(
        default=None,
        validator=_check_horizon,
        eq=False,
    )  # Exclude from __hash__
    #: Maximum of inverter in kilowatts or kVA; float > 0
    inverter: float | None = field(default=None, converter=float, validator=vld.gt(0.0), metadata={"QUERY_PARAM": True})
    #: Number of response days; int in (1,..,8), 1=today
    limit: int | None = field(default=None, validator=vld.in_(tuple(range(1, 9))), metadata={"QUERY_PARAM": True})
    #: Start time of the forecast, PHP DateTime format; string
    start: str | None = field(default=None, converter=str, validator=_check_php_datetime)
    #: Flag for including solar transit (midday)
    transit: int | None = field(default=None, converter=int, validator=vld.in_((0, 1)), metadata={"QUERY_PARAM": True})
    #: Actual production until now; float >= 0
    actual: float | None = field(default=None, converter=float, validator=vld.ge(0.0), metadata={"QUERY_PARAM": True})

    #: Url parameters for the API; dict
    _url_params: dict[str, Any] = field(init=False, repr=False, eq=False, order=False)
    #: Query parameters for the API; dict
    _query_params: dict[str, Any] = field(init=False, repr=False, eq=False, order=False)

    def __attrs_post_init__(self) -> None:
        """Process attributes after initialization."""
        if self.url not in [None, ""]:
            log.info("Passing 'url' to ForecastSolar node is not supported and will be ignored.")

        if not (isinstance(self.declination, int) and isinstance(self.azimuth, int) and isinstance(self.kwp, float)):
            if isinstance(self.declination, list) and isinstance(self.azimuth, list) and isinstance(self.kwp, list):
                if not len(self.declination) == len(self.azimuth) == len(self.kwp):
                    raise ValueError("Declination, azimuthimuth and kwp must be passed for all planes")
                if self.api_key is None:
                    raise ValueError("Valid API key is needed for multiple planes")
            else:
                raise ValueError("Declination, azimuth and kwp must be passed either as lists or as single values.")

        if self.api_key is None and (self.endpoint not in ["estimate", "check"]):
            raise ValueError(f"Valid API key is needed for endpoint: {self.endpoint}")
        if self.endpoint != "estimate" and self.data is not None:
            log.info("'data' field is not supported for endpoints other than estimate and will be ignored.")
            object.__setattr__(self, "data", None)
        elif self.endpoint == "estimate" and self.data is None:
            log.info("'data' field will be set to 'watts' for endpoint 'estimate'.")
            object.__setattr__(self, "data", "watts")

        # Collect all url parameters and query parameters
        url_params = {}
        query_params = {}
        for _field in self.__attrs_attrs__:  # type: ignore[attr-defined]
            if _field.name is not None:
                if _field.metadata.get("QUERY_PARAM") is False:
                    url_params[_field.name] = getattr(self, _field.name)
                elif _field.metadata.get("QUERY_PARAM") is True:
                    query_params[_field.name] = getattr(self, _field.name)

        # Construct the URL
        url = self._build_url(url_params)

        object.__setattr__(self, "url", url)
        object.__setattr__(self, "_url_params", url_params)
        object.__setattr__(self, "_query_params", query_params)

        super().__attrs_post_init__()

    def _build_url(self, url_params: dict[str, Any]) -> str:
        """Build the URL for the Forecast Solar API.

        :param url_params: dictionary with URL parameters.
        :return: URL for the Forecast Solar API.
        """
        url = "https://api.forecast.solar"
        keys = []

        # Check if the API key is set and add it to the URL
        if self.api_key is not None:
            keys.append("api_key")

        keys.append("endpoint")

        if self.endpoint == "estimate":
            keys.append("data")
        keys.extend(["latitude", "longitude"])

        for path in keys:
            try:
                url += f"/{url_params[path]}"
            except KeyError:
                continue

        # Unpack plane parameters and add them to the URL
        if isinstance(self.declination, list):
            for i in range(len(self.declination)):
                url += f"/{self.declination[i]}/{self.azimuth[i]}/{self.kwp[i]}"  # type: ignore[index]
        else:
            url += f"/{self.declination}/{self.azimuth}/{self.kwp}"

        return url

    @classmethod
    def _get_params(cls, dikt: dict[str, Any]) -> dict[str, Any]:
        """Get the common parameters for a Forecast Solar node.

        :param dikt: dictionary with node information.
        :return: dict with: api_key, endpoint, latitude, longitude, declination, azimuth, kwp
        """
        return {
            "endpoint": dikt.get("endpoint", "estimate"),
            "latitude": int(dikt.get("latitude", 0) or dikt.get("lat", 0)),
            "longitude": int(dikt.get("longitude", 0) or dikt.get("lon", 0)),
            "declination": dikt.get("declination", 0) or dikt.get("dec", 0),
            "azimuth": dikt.get("azimuth", 0) or dikt.get("az", 0),
            "kwp": dikt.get("kwp", 0),
        }

    @classmethod
    def _from_dict(cls, dikt: dict[str, Any]) -> NodeForecastSolar:
        """Create a Forecast Solar node from a dictionary of node information.

        :param dikt: dictionary with node information.
        :return: NodeForecastSolar object.
        """
        name, _, url, _, _ = cls._read_dict_info(dikt)

        params = cls._get_params(dikt)

        _data_point = str(dict_get_any(dikt, "data", fail=False))
        if _data_point not in ["None", "nan"]:
            params["data"] = _data_point

        dict_key = str(dict_get_any(dikt, "api_key", "apikey", fail=False))
        if dict_key not in ["None", "nan"]:
            params["api_key"] = dict_key
        else:
            log.info(
                """The api_key is None.
                Make sure to set an api_key to use the personal or the professional functions of forecastsolar.api
                otherwise the public functions are only available."""
            )

        for key in ["declination", "azimuth", "kwp"]:
            if isinstance(params[key], str):
                params[key] = ast.literal_eval(params[key])

        try:
            return cls(name, url, "forecast_solar", **params)
        except (TypeError, AttributeError) as e:
            raise TypeError(
                f"""Could not convert all types for node {name}:
                            \n{e}"""
            ) from e



from __future__ import annotations

import importlib
import inspect
import os
import pathlib
import subprocess
import sys
from shutil import which
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from types import ModuleType

    from eta_utility.type_hints import Path

# Set environment variable to determine the correct python environment to use when calling back into
# python from julia
os.environ["PYCALL_JL_RUNTIME_PYTHON"] = sys.executable

JULIA_NOT_FOUND_MSG = (
    "Julia executable cannot be found. "
    "If you have installed Julia, make "
    "sure Julia executable is in the system path. "
    "If you have not installed Julia, download from "
    "https://julialang.org/downloads/ and install it. "
)


def import_jl_file(filename: Path) -> ModuleType:
    check_julia_package()

    _filename = pathlib.Path(filename) if not isinstance(filename, pathlib.Path) else filename
    jl = importlib.import_module("julia.Main")

    jl.include(_filename.absolute().as_posix())
    return jl


def import_jl(importstr: str) -> ModuleType:
    """Import a julia file into the main namespace. The syntax is equivalent to python import strings. If the import
    string starts with a '.', the import path is interpreted as relative to the file calling this function. If the
    import string is absolute, it will use the python sys.path list to look for the file.

    The function also makes sure that julia.Main is imported and returns a handle to the module. This way, the
    imported julia file can be used right away.

    :param importstr: Path to the imported julia package. If the path starts with a '.' this will be relative to the
        file it is specified in. Otherwise, this will look through python import Path.
    """
    check_julia_package()

    file = importstr_to_path(importstr, _stack=2)
    return import_jl_file(file)


def importstr_to_path(importstr: str, _stack: int = 1) -> pathlib.Path:
    """Converts an import string into a python path. The syntax is equivalent to python import strings. If the
    import string starts with a '.', the import path is interpreted as relative to the file calling this function.
    If the import string is absolute, it will use the python sys.path list to look for the file.

    :param importstr: Path to the imported julia package (python import string). If the path starts with a '.' this
        will be relative to the file it is specified in. Otherwise, this will look through the python import paths.
    """
    if len(importstr) > 2 and importstr[0] == "." and importstr[1] == ".":
        pathstr = f"..{importstr[2:].replace('.', '/')}.jl"
        relative = True
    elif len(importstr) > 1 and importstr[0] == ".":
        pathstr = f"{importstr[1:].replace('.', '/')}.jl"
        relative = True
    else:
        pathstr = f"{importstr.replace('.', '/')}.jl"
        relative = False

    file = None
    found = False
    if relative:
        file = pathlib.Path(inspect.stack()[_stack].filename).parent / pathstr
        if file.is_file():
            found = True
    else:
        for path in sys.path:
            file = pathlib.Path(path) / pathstr
            if file.is_file():
                found = True
                break

    if not found and relative and file:
        raise ImportError(f"Could not find the specified julia file. Looking for {file}")
    if not found or not file:
        raise ImportError(f"Could not find the specified julia file. Looking for {pathstr}")

    return file


def update_agent() -> None:
    """Upadtes the NSGA2 agent model file"""
    import tempfile

    from test.test_etax.test_agents import TestNSGA2

    cls = TestNSGA2()
    cls.create_stored_agent_file("test/resources/agents/", tempfile.TemporaryDirectory().name)


def install_julia() -> None:
    """Checks if Julia language is available in the system and install and configure pyjulia.
    Also install ju_extensions in Julia environmnent.
    """
    if which("julia") is None:
        raise ImportError(JULIA_NOT_FOUND_MSG)

    try:
        import julia
    except ImportError:
        subprocess.check_call([sys.executable, "-m", "pip", "install", "julia"])

        import julia

    # Set environment variable to determine the correct python environment to use when calling back into
    # python from julia
    os.environ["PYCALL_JL_RUNTIME_PYTHON"] = sys.executable
    julia.install()

    subprocess.check_call(
        [
            "julia",
            "-e",
            f"import Pkg; Pkg.develop(path=\"{(pathlib.Path(__file__).parent / 'ju_extensions').as_posix()}\")",
        ]
    )


def check_julia_package() -> bool:
    """Check if everything is available and setup correctly to execute modules depending on eta_utility
    julia extensions. This function raises ImportError if necessary components are missing.

    :returns: True if is installed ImportError if not
    """
    if which("julia") is None:
        raise ImportError(JULIA_NOT_FOUND_MSG)

    try:
        import julia
    except ModuleNotFoundError:
        raise ImportError(
            "Could not find the python julia package. Please run the command: install-julia "
            "inside the python virtual environment where eta-utility is installed."
        ) from None

    try:
        from julia import ju_extensions  # noqa: F401
    except julia.core.UnsupportedPythonError as e:
        raise ImportError(
            "PyCall for Julia is installed for a different python binary than you are currently "
            "using. Please run the command: install-julia inside the python virtual environment "
            "where eta-utility is installed."
        ) from e
    except (ModuleNotFoundError, ImportError, AttributeError) as e:
        raise ImportError(
            "Could not find julia extension module for eta_utility (ju_extensions missing). Please "
            "run the command: install-julia inside the python virtual environment where eta-utility "
            "is installed."
        ) from e

    return True


def julia_extensions_available() -> bool:
    """Check if everything is available and setup correctly to execute modules depending on eta_utility
    julia extensions. This function returns false if necessary components are missing. It does not
    provide any indications what is missing.

    :return: True if julia extensions are correctly installed, false if not.
    """
    try:
        check_julia_package()
    except ImportError:
        return False

    return True



from __future__ import annotations

import asyncio
import pathlib
import time
from collections import deque
from datetime import timedelta
from typing import TYPE_CHECKING, NamedTuple

import numpy as np
import onnxruntime
import pandas as pd

from eta_utility import get_logger
from eta_utility.connectors import (
    DFSubHandler,
    OpcUaConnection,
    name_map_from_node_sequence,
)
from eta_utility.connectors.node import NodeOpcUa
from eta_utility.servers import OpcUaServer

if TYPE_CHECKING:
    from typing import Final

    from eta_utility.type_hints import Path, TimeStep

log = get_logger(level=1)


class Config:
    #: Inference interval.
    INTERVAL: Final[int] = 1
    #: Path to the trained inference model.
    PATH_MODEL: Final[pathlib.Path] = pathlib.Path(__file__).parent / "resources/emag_gt_forecast.onnx"
    #: Shape of the inputs provided to the model.
    INPUT_SHAPE: Final[list[int]] = [1, 5, 20, 9]

    #: Path to a csv file containing normalization data.
    PATH_NORMALIZATION_DATA: Final[pathlib.Path] = pathlib.Path(__file__).parent / "resources/features_max_min.csv"
    #: Names of features used for normalization.
    FEATURES_NORMALIZATION: Final[list[str]] = [
        "WZM_VLC-100GT.972.Elek_P.L1-4",
        "WZM_VLC-100GT.972.Elek_P.KSS",
        "WZM_VLC-100GT.972.Elek_P.Ax_C1",
        "WZM_VLC-100GT.972.Elek_P.Sp3",
        "WZM_VLC-100GT.972.Elek_P.Sp4",
        "WZM_VLC-100GT.972.Elek_P.Ax_Q1",
        "WZM_VLC-100GT.972.Elek_P.Ax_X1",
        "WZM_VLC-100GT.972.Elek_P.Ax_Z1",
        "regressand",
    ]

    #: OPC Server and nodes for the input features.
    OPC_SERVER: Final[dict[str, str | int]] = {"ip": "localhost", "port": 48050}
    #: Nodes to use as input features.
    NODES: Final[list[str]] = [
        "ns=2;s=Application.GbIL4EE.powerMain",
        "ns=2;s=Application.GbIL4EE.rpowerKSS_System_ND",
        "ns=2;s=Application.GbIL4EE.powerC1",
        "ns=2;s=Application.GbIL4EE.powerSP3",
        "ns=2;s=Application.GbIL4EE.powerSP4",
        "ns=2;s=Application.GbIL4EE.powerQ1",
        "ns=2;s=Application.GbIL4EE.powerX1",
        "ns=2;s=Application.GbIL4EE.powerZ1",
    ]
    LOCAL_FILE: Final[pathlib.Path] = pathlib.Path(__file__).parent / "resources/input_features_local.csv"

    #: Name of the node for publishing the resulting data.
    PUBLISH_NAME: Final[str] = "ns=6;s=EMAG-GT-forecast.electric_power_100s"


def main() -> None:
    asyncio.get_event_loop().create_task(local_server())
    forecasting()


def forecasting() -> None:
    """Initialize and execute the forecasting loop."""

    # load maximal and minimal values of the features for normalization
    features_max, features_min = load_normalization_params(
        Config.PATH_NORMALIZATION_DATA, Config.FEATURES_NORMALIZATION
    )

    # Create input connections
    connection = OpcUaConnection.from_ids(
        Config.NODES, f"opc.tcp://{Config.OPC_SERVER['ip']}:{Config.OPC_SERVER['port']}"
    )
    sub_handler = DFSubHandler(write_interval=Config.INTERVAL, size_limit=100, auto_fillna=False)
    connection.subscribe(sub_handler, interval=Config.INTERVAL)

    # Create loop and inference tasks
    loop = asyncio.get_event_loop()
    data_queue: asyncio.Queue = asyncio.Queue()

    running_tasks = set()
    read_data_task = loop.create_task(read_data_loop(sub_handler, data_queue, Config.INPUT_SHAPE[-1], Config.INTERVAL))
    running_tasks.add(read_data_task)
    inference_task = loop.create_task(
        inference_loop(
            Config.PATH_MODEL, data_queue, Config.PUBLISH_NAME, Config.INPUT_SHAPE, features_min, features_max
        )
    )
    running_tasks.add(inference_task)

    try:
        log.info("Starting processing loop.")
        loop.run_forever()
    except KeyboardInterrupt:
        loop.stop()
        log.warning("Keyboard Interrupt. Stopping inference.")


async def local_server() -> None:
    data = pd.read_csv(Config.LOCAL_FILE, sep=";", decimal=",")
    current_line = 0

    # initialize the server
    nodes = [
        NodeOpcUa(
            name=name,
            url=f"opc.tcp://{Config.OPC_SERVER['ip']}:{Config.OPC_SERVER['port']}",
            protocol="opcua",
            opc_id=name,
            dtype="float",
        )
        for name in data.columns
    ]
    server = OpcUaServer(namespace=2, ip=str(Config.OPC_SERVER["ip"]), port=int(Config.OPC_SERVER["port"]))
    server.create_nodes(nodes)
    node_map: dict[str, NodeOpcUa] = name_map_from_node_sequence(nodes)
    while True:
        server.write({node_map[name]: value for name, value in data.iloc[current_line].items()})
        current_line = current_line + 1 if current_line < len(data) else 0
        await asyncio.sleep(1)


def load_normalization_params(csv_file: Path, features: list[str]) -> tuple[list, list]:
    """Get minimum and maximum values of features for normalization and inverse normalization.

    :param csv_file: Path to csv file that includes min max values saved from past data.
    :param features: List of feature names used for normalization.
    :returns: List of max values, list of min values
    """

    min_max = pd.read_csv(csv_file, header=0, index_col=0)

    features_max = [min_max.get(feature).to_numpy()[0] for feature in features]
    features_min = [min_max.get(feature).to_numpy()[1] for feature in features]

    return features_max, features_min


async def read_data_loop(
    sub_handler: DFSubHandler, data_queue: asyncio.Queue, expected_values: int, interval: TimeStep
) -> None:
    """Read data from a data subscription and adjust the format to correspond to input shape requirements.

    :param sub_handler: Subscription Handler to read data from.
    :param data_queue: Queue to put the collected data into.
    :param expected_values: Number of values to expect from the subscription handler.
    :param interval: Interval for requesting data from the subscription.
    """
    _interval = interval.total_seconds() if isinstance(interval, timedelta) else interval

    class InputData(NamedTuple):
        date: str
        data: list[float]

    while True:
        start_time = time.time()

        async def sleep(_start_time: float) -> None:
            # Log timing and wait for next loop
            finish_time = time.time()
            log.debug(f"Time used in read loop: {finish_time - _start_time:.4f} s")
            await asyncio.sleep(_start_time + _interval - finish_time)

        data = sub_handler.get_latest()
        if data is None:
            await sleep(start_time)
            continue

        date = data.index[-1]
        adjusted_data = [np.nan_to_num(data.loc[date, col]) for col in data.columns]
        adjusted_data.append(sum(adjusted_data))  # Calculate and append the regressand

        if len(adjusted_data) != expected_values:
            await sleep(start_time)
            continue

        await data_queue.put(InputData(date.to_pydatetime(), adjusted_data))
        await sleep(start_time)


async def inference_loop(
    model_path: Path,
    data_queue: asyncio.Queue,
    publish_name: str,
    input_shape: list,
    features_min: list[float],
    features_max: list[float],
) -> None:
    """Asynchronous task for value inference using the ONNX runtime.

    :param model_path: Path to the stored ONNX inference model.
    :param data_queue: Queue to receive historical data.
    :param publish_name: Name of the OPC UA node for publishing the data.
    :param input_shape: Input shape expected by ONNX model.
    :param features_min: Minimum values of features, used for normalization of input features.
    :param features_max: Maximum values of features, used for normalization of input features.
    """
    maxlen = 100
    # initialize inference with onnxruntime
    session = onnxruntime.InferenceSession(str(model_path), onnxruntime.SessionOptions())
    log.info("Starting inference with ONNX-Runtime.")

    model_input_list: deque = deque(maxlen=maxlen)

    output_node = NodeOpcUa(
        name="EMAG-GT-forecast-opcua",
        url="opc.tcp://localhost",
        protocol="opcua",
        opc_id=publish_name,
    )
    with OpcUaServer(namespace=6, ip="localhost") as output_server:
        output_server.create_nodes(output_node)

        while True:
            data = await data_queue.get()
            start_time = time.time()

            # Normalize and append data and check whether enough historic data is available for inference
            model_input_list.append([(a - b) / (c - b) for a, b, c in zip(data.data, features_min, features_max)])
            if len(model_input_list) < maxlen:
                log.info(f"Not enough input data. Need {maxlen - len(model_input_list)} more lines.")
                continue

            # Perform the prediction
            inferred_value = session.run(
                None, {session.get_inputs()[0].name: np.array(model_input_list, dtype=np.float32).reshape(input_shape)}
            )[0][0][0]
            log.info(f"ONNX predicted value: {inferred_value}")

            # De-Normalized the prediction
            inverse_prediction = inferred_value * (features_max[-1] - features_min[-1]) + features_min[-1]
            log.info(f"Inverse predicted value: {inverse_prediction}")
            log.debug(f"Time used for inference: {time.time() - start_time:.4f} s")

            # Publish results to OPC UA server
            output_server.write({output_node: inverse_prediction})
            log.info("Value written to OPC UA server")


if __name__ == "__main__":
    main()



"""Utility functions for connecting to the EnEffCo database and reading data."""

from __future__ import annotations

import asyncio
import concurrent.futures
from collections.abc import Mapping
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import requests
from requests_cache import CachedSession

from eta_utility import get_logger
from eta_utility.connectors.node import NodeEnEffCo

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any

    from eta_utility.type_hints import Nodes, TimeStep

from .base_classes import SeriesConnection, SubscriptionHandler

log = get_logger("connectors.eneffco")


class EnEffCoConnection(SeriesConnection[NodeEnEffCo], protocol="eneffco"):
    """
    EnEffCoConnection is a class to download and upload multiple features from and to the EnEffCo database as
    timeseries.

    :param url: URL of the server with scheme (https://).
    :param usr: Username in EnEffco for login.
    :param pwd: Password in EnEffco for login.
    :param api_token: Token for API authentication.
    :param nodes: Nodes to select in connection.
    """

    API_PATH: str = "/API/v1.0"

    def __init__(
        self, url: str, usr: str | None, pwd: str | None, *, api_token: str, nodes: Nodes[NodeEnEffCo] | None = None
    ) -> None:
        url = url + self.API_PATH
        self._api_token: str = api_token
        super().__init__(url, usr, pwd, nodes=nodes)

        if self.usr is None:
            raise ValueError("Username must be provided for the EnEffCo connector.")
        if self.pwd is None:
            raise ValueError("Password must be provided for the EnEffCo connector.")

        self._node_ids: pd.DataFrame | None = None
        self._node_ids_raw: pd.DataFrame | None = None

        self._sub: asyncio.Task | None = None
        self._subscription_nodes: set[NodeEnEffCo] = set()
        self._subscription_open: bool = False
        self._session: CachedSession = CachedSession(
            cache_name="eta_utility/connectors/requests_cache/eneffco_cache",
            expire_after=timedelta(minutes=15),
            use_cache_dir=True,
        )

    @classmethod
    def _from_node(
        cls, node: NodeEnEffCo, usr: str | None = None, pwd: str | None = None, **kwargs: Any
    ) -> EnEffCoConnection:
        """Initialize the connection object from an EnEffCo protocol node object

        :param node: Node to initialize from.
        :param usr: Username to use.
        :param pwd: Password to use.
        :param kwargs: Keyword arguments for API authentication, where "api_token" is required
        :return: EnEffCoConnection object.
        """

        api_token = kwargs.get("api_token")
        if api_token is None:
            raise AttributeError("Keyword parameter 'api_token' is missing.")

        return super()._from_node(node, usr=usr, pwd=pwd, api_token=api_token)

    @classmethod
    def from_ids(cls, ids: Sequence[str], url: str, usr: str, pwd: str, api_token: str) -> EnEffCoConnection:
        """Initialize the connection object from an EnEffCo protocol through the node IDs

        :param ids: Identification of the Node.
        :param url: URL for EnEffco connection.
        :param usr: Username for EnEffCo login.
        :param pwd: Password for EnEffCo login.
        :param api_token: Token for API authentication.
        :return: EnEffCoConnection object.
        """
        nodes = [NodeEnEffCo(name=name, url=url, protocol="eneffco", eneffco_code=name) for name in ids]
        return cls(url=url, usr=usr, pwd=pwd, api_token=api_token, nodes=nodes)

    def read(self, nodes: Nodes[NodeEnEffCo] | None = None) -> pd.DataFrame:
        """Download current value from the EnEffCo Database

        :param nodes: List of nodes to read values from.
        :return: pandas.DataFrame containing the data read from the connection.
        """
        nodes = self._validate_nodes(nodes)
        base_time = 1  # seconds
        the_time = self._round_timestamp(datetime.now(), base_time).replace(tzinfo=None)
        return self.read_series(the_time - timedelta(seconds=base_time), the_time, nodes, base_time)

    def write(
        self, values: Mapping[NodeEnEffCo, Any] | pd.Series[datetime, Any], time_interval: timedelta | None = None
    ) -> None:
        """Writes some values to the EnEffCo Database

        :param values: Dictionary of nodes and data to write {node: value}.
        :param time_interval: Interval between datapoints (i.e. between "From" and "To" in EnEffCo Upload) (default 1s).
        """
        nodes = self._validate_nodes(list(values.keys()))

        if time_interval is None:
            time_interval = timedelta(seconds=1)

        for node in nodes:
            request_url = f"rawdatapoint/{self.id_from_code(node.eneffco_code, raw_datapoint=True)}/value"
            response = self._raw_request(
                "POST",
                request_url,
                data=self._prepare_raw_data(values[node], time_interval),
                headers={
                    "Content-Type": "application/json",
                    "cache-control": "no-cache",
                    "Postman-Token": self._api_token,
                },
                params={"comment": ""},
            )
            log.info(response.text)

    def _prepare_raw_data(
        self, data: Mapping[datetime, Any] | pd.Series[datetime, Any], time_interval: timedelta
    ) -> str:
        """Change the input format into a compatible format with EnEffCo and filter NaN values.

        :param data: Data to write to node {time: value}. Could be a dictionary or a pandas Series.
        :param time_interval: Interval between datapoints (i.e. between "From" and "To" in EnEffCo Upload).

        :return upload_data: String from dictionary in the format for the upload to EnEffCo.
        """

        if isinstance(data, (dict, pd.Series)):
            upload_data: dict[str, list[Any]] = {"Values": []}
            for time, val in data.items():
                # Only write values if they are not nan
                if not np.isnan(val):
                    aware_time = self._assert_tz_awareness(time).astimezone(timezone.utc)
                    upload_data["Values"].append(
                        {
                            "Value": float(val),
                            "From": aware_time.strftime("%Y-%m-%d %H:%M:%SZ"),
                            "To": (aware_time + time_interval).strftime("%Y-%m-%d %H:%M:%SZ"),
                        }
                    )

        else:
            raise ValueError("Unrecognized data format for EnEffCo upload. Provide dictionary or pandas series.")

        return str(upload_data)

    def read_info(self, nodes: Nodes[NodeEnEffCo] | None = None) -> pd.DataFrame:
        """Read additional datapoint information from Database.

        :param nodes: List of nodes to read values from.
        :return: pandas.DataFrame containing the data read from the connection.
        """

        nodes = self._validate_nodes(nodes)
        values = []

        for node in nodes:
            request_url = f"datapoint/{self.id_from_code(node.eneffco_code)}"
            response = self._raw_request("GET", request_url)
            values.append(pd.Series(response.json(), name=node.name))

        return pd.concat(values, axis=1)

    def subscribe(
        self, handler: SubscriptionHandler, nodes: Nodes[NodeEnEffCo] | None = None, interval: TimeStep = 1
    ) -> None:
        """Subscribe to nodes and call handler when new data is available. This will return only the
        last available values.

        :param handler: SubscriptionHandler object with a push method that accepts node, value pairs.
        :param interval: Interval for receiving new data. It is interpreted as seconds when given as an integer.
        :param nodes: Identifiers for the nodes to subscribe to.
        """
        self.subscribe_series(handler=handler, req_interval=1, nodes=nodes, interval=interval, data_interval=interval)

    def read_series(
        self,
        from_time: datetime,
        to_time: datetime,
        nodes: Nodes[NodeEnEffCo] | None = None,
        interval: TimeStep = 1,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """Download timeseries data from the EnEffCo Database

        :param nodes: List of nodes to read values from.
        :param from_time: Starting time to begin reading (included in output).
        :param to_time: Time to stop reading at (not included in output).
        :param interval: Interval between time steps. It is interpreted as seconds if given as integer.
        :param kwargs: Other parameters (ignored by this connector).
        :return: Pandas DataFrame containing the data read from the connection.
        """
        _interval = interval if isinstance(interval, timedelta) else timedelta(seconds=interval)

        nodes = self._validate_nodes(nodes)

        def read_node(node: NodeEnEffCo) -> pd.DataFrame:
            request_url = (
                f"datapoint/{self.id_from_code(node.eneffco_code)}/value?"
                f"from={self.timestr_from_datetime(from_time)}&"
                f"to={self.timestr_from_datetime(to_time)}&"
                f"timeInterval={int(_interval.total_seconds())!s}&"
                "includeNanValues=True"
            )
            response = self._raw_request("GET", request_url).json()

            data = pd.DataFrame(
                data=(r["Value"] for r in response),
                index=pd.to_datetime([r["From"] for r in response], utc=True, format="%Y-%m-%dT%H:%M:%SZ").tz_convert(
                    self._local_tz
                ),
                columns=[node.name],
                dtype="float64",
            )
            data.index.name = "Time (with timezone)"
            return data

        with concurrent.futures.ThreadPoolExecutor() as executor:
            results = executor.map(read_node, nodes)

        return pd.concat(results, axis=1, sort=False)

    def subscribe_series(
        self,
        handler: SubscriptionHandler,
        req_interval: TimeStep,
        offset: TimeStep | None = None,
        nodes: Nodes[NodeEnEffCo] | None = None,
        interval: TimeStep = 1,
        data_interval: TimeStep = 1,
        **kwargs: Any,
    ) -> None:
        """Subscribe to nodes and call handler when new data is available. This will always return a series of values.
        If nodes with different intervals should be subscribed, multiple connection objects are needed.

        :param handler: SubscriptionHandler object with a push method that accepts node, value pairs.
        :param req_interval: Duration covered by requested data (time interval). Interpreted as seconds if given as int.
        :param offset: Offset from datetime.now from which to start requesting data (time interval).
                       Interpreted as seconds if given as int. Use negative values to go to past timestamps.
        :param data_interval: Time interval between values in returned data. Interpreted as seconds if given as int.
        :param interval: interval (between requests) for receiving new data.
                         It is interpreted as seconds when given as an integer.
        :param nodes: Identifiers for the nodes to subscribe to.
        :param kwargs: Other, ignored parameters.
        """

        nodes = self._validate_nodes(nodes)

        interval = interval if isinstance(interval, timedelta) else timedelta(seconds=interval)
        req_interval = req_interval if isinstance(req_interval, timedelta) else timedelta(seconds=req_interval)
        if offset is None:
            offset = -req_interval
        else:
            offset = offset if isinstance(offset, timedelta) else timedelta(seconds=offset)
        data_interval = data_interval if isinstance(data_interval, timedelta) else timedelta(seconds=data_interval)

        self._subscription_nodes.update(nodes)

        if self._subscription_open:
            # Adding nodes to subscription is enough to include them in the query. Do not start an additional loop
            # if one already exists
            return

        self._subscription_open = True
        loop = asyncio.get_event_loop()
        self._sub = loop.create_task(
            self._subscription_loop(
                handler,
                int(interval.total_seconds()),
                req_interval,
                offset,
                data_interval,
            )
        )

    def close_sub(self) -> None:
        """Close an open subscription."""
        self._subscription_open = False

        if self.exc:
            raise self.exc

        try:
            self._sub.cancel()  # type: ignore
        except Exception:
            pass

    async def _subscription_loop(
        self,
        handler: SubscriptionHandler,
        interval: TimeStep,
        req_interval: TimeStep,
        offset: TimeStep,
        data_interval: TimeStep,
    ) -> None:
        """The subscription loop handles requesting data from the server in the specified interval

        :param handler: Handler object with a push function to receive data.
        :param interval: Interval for requesting data in seconds.
        :param req_interval: Duration covered by the requested data.
        :param offset: Offset from datetime.now from which to start requesting data (time interval).
                       Use negative values to go to past timestamps.
        :param data_interval: Interval between data points.
        """
        interval = interval if isinstance(interval, timedelta) else timedelta(seconds=interval)
        req_interval = req_interval if isinstance(req_interval, timedelta) else timedelta(seconds=req_interval)
        data_interval = data_interval if isinstance(data_interval, timedelta) else timedelta(seconds=data_interval)
        offset = offset if isinstance(offset, timedelta) else timedelta(seconds=offset)

        try:
            while self._subscription_open:
                from_time = datetime.now() + offset
                to_time = from_time + req_interval

                values = self.read_series(from_time, to_time, self._subscription_nodes, interval=data_interval)
                for node in self._subscription_nodes:
                    handler.push(node, values[node.name])

                await asyncio.sleep(interval.total_seconds())
        except BaseException as e:
            self.exc = e

    def id_from_code(self, code: str, raw_datapoint: bool = False) -> str:
        """
        Function to get the raw EnEffCo ID corresponding to a specific (raw) datapoint

        :param code: Exact EnEffCo code.
        :param raw_datapoint: Returns raw datapoint ID.
        """

        # Only build lists of IDs if they are not available yet
        if self._node_ids is None:
            response = self._raw_request("GET", "/datapoint")
            self._node_ids = pd.DataFrame(data=response.json())

        if self._node_ids_raw is None:
            response = self._raw_request("GET", "/rawdatapoint")
            self._node_ids_raw = pd.DataFrame(data=response.json())

        def find_id(node_ids: pd.DataFrame) -> str:
            if len(node_ids.loc[node_ids["Code"] == code, "Id"]) <= 0:
                raise ValueError(f"Code {code} does not exist on server {self.url}.")
            return node_ids.loc[node_ids["Code"] == code, "Id"].to_numpy().item()

        return find_id(self._node_ids_raw) if raw_datapoint else find_id(self._node_ids)

    def timestr_from_datetime(self, dt: datetime) -> str:
        """Create an EnEffCo compatible time string.

        :param dt: Datetime object to convert to string.
        :return: EnEffCo compatible time string.
        """

        return dt.isoformat(sep="T", timespec="seconds").replace(":", "%3A").replace("+", "%2B")

    def _raw_request(self, method: str, endpoint: str, **kwargs: Any) -> requests.Response:
        """Perform EnEffCo request and handle possibly resulting errors.

        :param method: HTTP request method.
        :param endpoint: Endpoint for the request (server URI is added automatically).
        :param kwargs: Additional arguments for the request.
        """
        assert self.usr is not None, "Make sure to specify a username before performing EnEffCo requests."
        assert self.pwd is not None, "Make sure to specify a password before performing EnEffCo requests."

        response = self._session.request(
            method, self.url + "/" + str(endpoint), auth=requests.auth.HTTPBasicAuth(self.usr, self.pwd), **kwargs
        )
        response.raise_for_status()

        return response

    def _validate_nodes(self, nodes: Nodes[NodeEnEffCo] | None) -> set[NodeEnEffCo]:  # type: ignore
        vnodes = super()._validate_nodes(nodes)
        _nodes = set()
        for node in vnodes:
            if isinstance(node, NodeEnEffCo):
                _nodes.add(node)

        return _nodes



from __future__ import annotations

import abc
import inspect
import json
import pathlib
from functools import partial
from typing import TYPE_CHECKING

import torch as th
from attrs import asdict
from stable_baselines3.common.vec_env import DummyVecEnv, VecMonitor, VecNormalize

from eta_utility import dict_get_any, get_logger, log_add_filehandler

from . import processors
from .policies import NoPolicy

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from typing import Any, Callable

    from gymnasium import Env
    from stable_baselines3.common.base_class import BaseAlgorithm, BasePolicy
    from stable_baselines3.common.vec_env import VecEnv

    from eta_utility.eta_x import ConfigOpt, ConfigOptRun
    from eta_utility.eta_x.envs import BaseEnv
    from eta_utility.type_hints import AlgoSettings, EnvSettings, Path

log = get_logger("eta_x")


def vectorize_environment(
    env: type[BaseEnv],
    config_run: ConfigOptRun,
    env_settings: EnvSettings,
    callback: Callable[[BaseEnv], None],
    verbose: int = 2,
    vectorizer: type[DummyVecEnv] = DummyVecEnv,
    n: int = 1,
    *,
    training: bool = False,
    monitor_wrapper: bool = False,
    norm_wrapper_obs: bool = False,
    norm_wrapper_reward: bool = False,
) -> VecNormalize | VecEnv:
    """Vectorize the environment and automatically apply normalization wrappers if configured. If the environment
    is initialized as an interaction_env it will not have normalization wrappers and use the appropriate configuration
    automatically.

    :param env: Environment class which will be instantiated and vectorized.
    :param config_run: Configuration for a specific optimization run.
    :param env_settings: Configuration settings dictionary for the environment which is being initialized.
    :param callback: Callback to call with an environment instance.
    :param verbose: Logging verbosity to use in the environment.
    :param vectorizer: Vectorizer class to use for vectorizing the environments.
    :param n: Number of vectorized environments to create.
    :param training: Flag to identify whether the environment should be initialized for training or playing. If true,
                     it will be initialized for training.
    :param norm_wrapper_obs: Flag to determine whether observations from the environments should be normalized.
    :param norm_wrapper_reward: Flag to determine whether rewards from the environments should be normalized.
    :return: Vectorized environments, possibly also wrapped in a normalizer.
    """
    # Create the vectorized environment
    log.debug("Trying to vectorize the environment.")
    # Ensure n is one, if the DummyVecEnv is used (it doesn't support more than one)
    if vectorizer.__class__.__name__ == "DummyVecEnv" and n != 1:
        n = 1
        log.warning("Setting number of environments to 1 because DummyVecEnv (default) is used.")

    if "verbose" in env_settings and env_settings["verbose"] is not None:
        verbose = env_settings.pop("verbose")

    # Create the vectorized environment
    def create_env(env_id: int) -> Env:
        env_id += 1
        return env(env_id=env_id, config_run=config_run, verbose=verbose, callback=callback, **env_settings)

    envs: VecEnv | VecNormalize
    envs = vectorizer([partial(create_env, i) for i in range(n)])

    # The VecMonitor knows the ep_reward and so this can be logged to tensorboard
    if monitor_wrapper:
        envs = VecMonitor(envs)

    # Automatically normalize the input features
    if norm_wrapper_obs or norm_wrapper_reward:
        # check if normalization data is available and load it if possible, otherwise
        # create a new normalization wrapper.
        if config_run.path_vec_normalize.is_file():
            log.info(
                f"Normalization data detected. Loading running averages into normalization wrapper: \n"
                f"\t {config_run.path_vec_normalize}"
            )
            envs = VecNormalize.load(str(config_run.path_vec_normalize), envs)
            envs.training = training
            envs.norm_obs = norm_wrapper_obs
            envs.norm_reward = norm_wrapper_reward
        else:
            log.info("No Normalization data detected.")
            envs = VecNormalize(envs, training=training, norm_obs=norm_wrapper_obs, norm_reward=norm_wrapper_reward)

    return envs


def initialize_model(
    algo: type[BaseAlgorithm],
    policy: type[BasePolicy],
    envs: VecEnv | VecNormalize,
    algo_settings: AlgoSettings,
    seed: int | None = None,
    *,
    tensorboard_log: bool = False,
    log_path: Path | None = None,
) -> BaseAlgorithm:
    """Initialize a new model or algorithm.

    :param algo: Algorithm to initialize.
    :param policy: The policy that should be used by the algorithm.
    :param envs: The environment which the algorithm operates on.
    :param algo_settings: Additional settings for the algorithm.
    :param seed: Random seed to be used by the algorithm.
    :param tensorboard_log: Flag to enable logging to tensorboard.
    :param log_path: Path for tensorboard log. Online required if logging is true
    :return: Initialized model.
    """
    log.debug(f"Trying to initialize model: {algo.__name__}")
    _log_path = log_path if log_path is None or isinstance(log_path, pathlib.Path) else pathlib.Path(log_path)

    # tensorboard logging
    algo_kwargs = {}
    if tensorboard_log:
        if _log_path is None:
            raise ValueError("If tensorboard logging is enabled, a path for results must be specified as well.")
        log.info(f"Tensorboard logging is enabled. Log file: {_log_path}")
        log.info(
            f"Please run the following command in the console to start tensorboard: \n"
            f'tensorboard --logdir "{_log_path}" --port 6006'
        )
        algo_kwargs = {"tensorboard_log": str(_log_path)}

    # check if the agent takes all the default parameters.
    algo_settings.setdefault("seed", seed)

    algo_params = inspect.signature(algo).parameters
    if "seed" not in algo_params and inspect.Parameter.VAR_KEYWORD not in {p.kind for p in algo_params.values()}:
        del algo_settings["seed"]
        log.warning(
            f"'seed' is not a valid parameter for agent {algo.__name__}. This default parameter will be ignored."
        )

    # create model instance
    return algo(policy, envs, **algo_settings, **algo_kwargs)  # type: ignore


def load_model(
    algo: type[BaseAlgorithm],
    envs: VecEnv | VecNormalize,
    algo_settings: AlgoSettings,
    path_model: Path,
    *,
    tensorboard_log: bool = False,
    log_path: Path | None = None,
) -> BaseAlgorithm:
    """Load an existing model.

    :param algo: Algorithm type of the model to be loaded.
    :param envs: The environment which the algorithm operates on.
    :param algo_settings: Additional settings for the algorithm.
    :param path_model: Path to load the model from.
    :param tensorboard_log: Flag to enable logging to tensorboard.
    :param log_path: Path for tensorboard log. Online required if logging is true
    :return: Initialized model.
    """
    log.debug(f"Trying to load existing model: {path_model}")
    _path_model = path_model if isinstance(path_model, pathlib.Path) else pathlib.Path(path_model)
    _log_path = log_path if log_path is None or isinstance(log_path, pathlib.Path) else pathlib.Path(log_path)

    if not _path_model.exists():
        raise OSError(f"Model couldn't be loaded. Path not found: {_path_model}")

    # tensorboard logging
    algo_kwargs = {}
    if tensorboard_log:
        if _log_path is None:
            raise ValueError("If tensorboard logging is enabled, a path for results must be specified as well.")
        log.info(f"Tensorboard logging is enabled. Log file: {_log_path}")
        log.info(
            f"Please run the following command in the console to start tensorboard: \n"
            f"tensorboard --logdir '{_log_path}' --port 6006"
        )
        algo_kwargs = {"tensorboard_log": str(_log_path)}

    try:
        model = algo.load(_path_model, envs, **algo_settings, **algo_kwargs)  # type: ignore
        log.debug("Model loaded successfully.")
    except OSError as e:
        raise OSError(f"Model couldn't be loaded: {e.strerror}. Filename: {e.filename}") from e

    return model


def log_to_file(config: ConfigOpt, config_run: ConfigOptRun) -> None:
    """Log output in terminal to the run_info file.

    :param config: Configuration to figure out the logging settings.
    :param config_run: Configuration for this optimization run.
    """

    file_path = config_run.path_log_output

    if config.settings.log_to_file:
        try:
            log_add_filehandler(filename=file_path)
        except Exception:
            log.error("Log file could not be created.")


def log_run_info(config: ConfigOpt, config_run: ConfigOptRun) -> None:
    """Save run configuration to the run_info file.

    :param config: Configuration for the framework.
    :param config_run: Configuration for this optimization run.
    """
    with config_run.path_run_info.open("w") as f:

        class Encoder(json.JSONEncoder):
            def default(self, o: object) -> object:
                if isinstance(o, pathlib.Path):
                    return str(o)
                if isinstance(o, abc.ABCMeta):
                    return None
                return repr(o)

        try:
            json.dump({**asdict(config_run), **asdict(config)}, f, indent=4, cls=Encoder)
            log.info("Log file successfully created.")
        except TypeError:
            log.warning("Log file could not be created because of non-serializable input in config.")


def deserialize_net_arch(
    net_arch: Sequence[Mapping[str, Any]], in_features: int, device: th.device | str = "auto"
) -> th.nn.Sequential:
    """Deserialize_net_arch can take a list of dictionaries describing a sequential torch network and deserialize
    it by instatiating the corresponding classes.

    An example for a possible net_arch would be:

    .. code-block::

        [{"layer": "Linear", "out_features": 60},
         {"activation_func": "Tanh"},
         {"layer": "Linear", "out_features": 60},
         {"activation_func": "Tanh"}]

    One key of the dictionary should be either 'layer', 'activation_func' or 'process'. If the 'layer' key is present,
    a layer from the :py:mod:`torch.nn` module is instantiated, if the 'activation_func' key is present, the
    value will be instantiated as an activation function from :py:mod:`torch.nn`. If the key 'process' is present,
    the value will be interpreted as a data processor from :py:mod:`eta_utility.eta_x.common.processors`.

    All other keys of each dictionary will be used as keyword parameters to the instantiation of the layer,
    activation function or processor.

    Only the number of input features for the first layer must be specified (using the 'in_features') parameter.
    The function will then automatically determine the number of input features for all other layers in the
    sequential network.

    :param net_arch: List of dictionaries describing the network architecture.
    :param in_features: Number of input features for the first layer.
    :param device: Torch device to use for training the network.
    :return: Sequential torch network.
    """
    network = th.nn.Sequential()
    _features = in_features

    for net in net_arch:
        _net = dict(net)
        if "process" in net:
            process = getattr(processors, _net.pop("process"))

            # The "Split" process must be treated differently, because it needs to be deserialized recursively.
            if {"net_arch" and "sizes"} < inspect.signature(process).parameters.keys():
                sizes = process.get_full_sizes(_features, _net["sizes"])
                _net["net_arch"] = [deserialize_net_arch(e, sizes[i], device) for i, e in enumerate(_net["net_arch"])]

            try:
                if len({"in_channels", "in_features"} & inspect.signature(process).parameters.keys()) > 0:
                    network.append(process(_features, **_net))
                else:
                    network.append(process(**_net))
            except TypeError as e:
                raise TypeError(f"Could not instantiate processing module {process.__name__}: {e}") from e

        elif "layer" in net:
            layer = getattr(th.nn, _net.pop("layer"))

            # Set the number of input features if required by the layer class
            try:
                if len({"in_channels", "in_features"} & inspect.signature(layer).parameters.keys()) > 0:
                    network.append(layer(_features, **_net))
                else:
                    network.append(layer(**_net))
            except TypeError as e:
                raise TypeError(f"Could not instantiate layer module {layer.__name__}: {e}") from e

        elif "activation_func" in net:
            activation_func = _net.pop("activation_func")
            try:
                network.append(getattr(th.nn, activation_func)(**_net))
            except TypeError as e:
                raise TypeError(f"Could not instantiate activation function module {activation_func}: {e}") from e
        else:
            raise ValueError(f"Unknown process or layer type: {net}.")

        _features = dict_get_any(_net, "out_channels", "out_features", fail=False, default=_features)

    network.to(device)
    return network


def log_net_arch(model: BaseAlgorithm, config_run: ConfigOptRun) -> None:
    """Store network architecture or policy information in a file. This requires for the model to be initialized,
    otherwise it will raise a ValueError.

    :param model: The algorithm whose network architecture is stored.
    :param config_run: Optimization run configuration (which contains info about the file to store info in).
    :raises: ValueError.
    """
    if not config_run.path_net_arch.exists() and model.policy is not None and model.policy.__class__ is not NoPolicy:
        with pathlib.Path(config_run.path_net_arch).open("w") as f:
            f.write(str(model.policy))

        log.info(f"Net arch / Policy information store successfully in: {config_run.path_net_arch}.")
    elif config_run.path_net_arch.exists():
        log.info(f"Net arch / Policy information already exists in {config_run.path_net_arch}")


def is_vectorized_env(env: BaseEnv | VecEnv | VecNormalize | None) -> bool:
    """Check if an environment is vectorized.

    :param env: The environment to check.
    """
    if env is None:
        return False

    return hasattr(env, "num_envs")


def is_env_closed(env: BaseEnv | VecEnv | VecNormalize | None) -> bool:
    """Check whether an environment has been closed.

    :param env: The environment to check.
    """
    if env is None:
        return True

    if hasattr(env, "closed"):
        return env.closed

    if hasattr(env, "venv"):
        return is_env_closed(env.venv)

    return False


def episode_results_path(series_results_path: Path, run_name: str, episode: int, env_id: int = 1) -> pathlib.Path:
    """Generate a filepath which can be used for storing episode results of a specific environment as a csv file.

    Name is of the format: ThisRun_001_01.csv (run name _ episode number _ environment id .csv)

    :param series_results_path: Path for results of the series of optimization runs.
    :param run_name: Name of the optimization run.
    :param episode: Number of the episode the environment is working on.
    :param env_id: Identification of the environment.
    """
    path = series_results_path if isinstance(series_results_path, pathlib.Path) else pathlib.Path(series_results_path)

    return path / f"{episode_name_string(run_name, episode, env_id)}.csv"


def episode_name_string(run_name: str, episode: int, env_id: int = 1) -> str:
    """Generate a name which can be used to pre or postfix files from a specific episode and run of an environment.

    Name is of the format: ThisRun_001_01 (run name _ episode number _ environment id)

    :param run_name: Name of the optimization run.
    :param episode: Number of the episode the environment is working on.
    :param env_id: Identification of the environment.
    """
    return f"{run_name}_{episode:0>#3}_{env_id:0>#2}"



from __future__ import annotations

import pathlib
from typing import TYPE_CHECKING

from eta_utility import get_logger
from eta_utility.connectors.node import NodeOpcUa
from eta_utility.eta_x import ETAx
from eta_utility.servers import OpcUaServer

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from typing import Any


def main() -> None:
    nodes = {
        "temp_tank": {"id": "ns=2;s=Application.general.Tanktemperatur.senT", "dtype": "float"},
        "tankheater": {"id": "ns=2;s=Application.outputs.Tankheizung", "dtype": "bool"},
        "mode_tankheater": {"id": "ns=2;s=Application.IL_4EEApplication.bTankheizungONOFF", "dtype": "bool"},
        "heating_register": {"id": "ns=2;s=Application.general.Scada_Heissblasen", "dtype": "bool"},
        "motor_nozzles": {"id": "ns=2;s=Application.outputs.Antrieb_Duesen_Freigabe", "dtype": "bool"},
        "motor_basket": {"id": "ns=2;s=Application.outputs.Antrieb_Korb_Freigabe", "dtype": "bool"},
        "pump": {"id": "ns=2;s=Application.outputs.Spritzpumpe", "dtype": "bool"},
        "fan": {"id": "ns=2;s=Application.outputs.Abluftgeblaese", "dtype": "bool"},
        "valve": {"id": "ns=2;s=Application.general.Scada_Impulsblasen", "dtype": "bool"},
    }

    get_logger(log_format="logname")
    local_server(local_nodes(nodes))

    root_path = get_path()
    experiment(root_path)


def experiment(root_path: pathlib.Path, overwrite: dict[str, Any] | None = None) -> None:
    """Perform a conventionally controlled experiment with the cleaning machine environment.

    :param root_path: Root path of the experiment.
    :param overwrite: Additional config values to overwrite values from JSON.
    """
    experiment = ETAx(
        root_path=root_path, config_overwrite=overwrite, relpath_config=".", config_name="experiment_cleaning_machine"
    )
    experiment.play("cleaning_machine", "test_run")


def local_nodes(definitions: Mapping[str, Mapping[str, str]]) -> list[NodeOpcUa]:
    """Create the specified list of local nodes.

    :param definitions: Node definition mapping
    :return: The created nodes
    """
    nodes = []

    for name, node in definitions.items():
        nodes.append(
            NodeOpcUa(
                name=name, url="opc.tcp://localhost:4840", protocol="opcua", opc_id=node["id"], dtype=node["dtype"]
            )
        )

    return nodes


def local_server(nodes: Sequence[NodeOpcUa]) -> OpcUaServer:
    """Create a local server with the specified nodes.

    :param Nodes: Sequence of nodes
    """
    server = OpcUaServer(namespace=9, ip="localhost")
    server.create_nodes(nodes)
    return server


def get_path() -> pathlib.Path:
    """Get the path of this file.

    :return: Path to this file.
    """
    return pathlib.Path(__file__).parent


if __name__ == "__main__":
    import sys

    sys.path.append(pathlib.Path(__file__).parent.parent.parent.as_posix())

    main()



from eta_utility.util_julia import julia_extensions_available

from .math_solver import (
    MathSolver as MathSolver,
    MPCBasic as MPCBasic,
)
from .rule_based import RuleBased as RuleBased

# Import Nsga2 algorithm if julia is available and ignore errors otherwise.
if julia_extensions_available():
    from .nsga2 import Nsga2 as Nsga2



from datetime import datetime

import pandas as pd

from eta_utility.connectors import ENTSOEConnection
from eta_utility.connectors.node import NodeEntsoE


def main() -> None:
    read_series()


def read_series() -> pd.DataFrame:
    # --begin_entsoe_doc_example--
    # Define your ENTSO-E Token
    entsoe_token = ""

    # Check out NodeEntsoE documentation for endpoint and bidding zone information
    node = NodeEntsoE(
        "CH1.Elek_U.L1-N",
        "https://transparency.entsoe.eu/",
        "entsoe",
        endpoint="Price",
        bidding_zone="DEU-LUX",
    )

    # start connection from one or multiple nodes
    server = ENTSOEConnection.from_node(node, api_token=entsoe_token)

    # Define time interval as datetime values
    from_datetime = datetime.strptime("2022-02-15T13:18:12", "%Y-%m-%dT%H:%M:%S")
    to_datetime = datetime.strptime("2022-02-15T14:00:00", "%Y-%m-%dT%H:%M:%S")

    # read_series will request data from specified connection and time interval
    # The DataFrame will have index with time delta of the specified interval in seconds
    if isinstance(server, ENTSOEConnection):
        result = server.read_series(from_time=from_datetime, to_time=to_datetime, interval=1)
    else:
        raise TypeError("The connection must be an ENTSOEConnection, to be able to call read_series.")
    # --end_entsoe_doc_example--

    return result


if __name__ == "__main__":
    main()



"""The script can be used to read data from different servers in the ETA-Factory.
It can write output to CSV files and/or publish to a different OPC UA server.
"""

from __future__ import annotations

import argparse
import asyncio
import pathlib
from datetime import timedelta
from typing import TYPE_CHECKING

from eta_utility.connectors.base_classes import Connection

try:
    import keyboard
except ModuleNotFoundError as e:
    raise ModuleNotFoundError(
        "For the data_recorder example, the pygame module is required. Install eta_utility with the "
        "[examples] option to get all packages required for running examples.",
        name="keyboard",
    ) from e
from eta_utility import get_logger
from eta_utility.connectors import Node, sub_handlers

if TYPE_CHECKING:
    from eta_utility.type_hints import Path, TimeStep


log = get_logger(level=2, log_format="logname")


def parse_args() -> argparse.Namespace:
    """Parse command line arguments (see help for description)."""
    parser = argparse.ArgumentParser()
    parser.add_argument("nodes_file", action="store", type=str, help="Excel file to read nodes from.")
    parser.add_argument("nodes_sheet", action="store", type=str, help="Name of the Excel sheet specifying the nodes.")
    parser.add_argument("--output_file", action="store", type=str, default=None, help="Path to the CSV output file.")
    parser.add_argument("--stop_after", action="store", type=int, default=None, help="Stop recording after X seconds.")
    parser.add_argument(
        "--sub_interval",
        action="store",
        type=float,
        default=1,
        help="Subscript read interval (or polling interval) in seconds.",
    )
    parser.add_argument(
        "--write_interval",
        action="store",
        type=float,
        default=1,
        help="Writing interval in seconds for writing to CSV file.",
    )
    parser.add_argument("--eneffco_usr", action="store", type=str, default=None, help="EnEffCo user name.")
    parser.add_argument("--eneffco_pw", action="store", type=str, default=None, help="EnEffCo password.")
    parser.add_argument(
        "--verbosity",
        action="store",
        type=int,
        default=2,
        help="Verbosity level (between 0 - no output and 4 - debug).",
    )

    return parser.parse_args()


async def logger(interval: int) -> None:
    """Print info message every interval seconds to show that the program continues to work.

    :param interval: Interval for printing the message in seconds.
    """
    step = 0
    while True:
        await asyncio.sleep(interval)
        step += interval
        log.info(f"Logging data for {step} s")


async def stop_execution(sleep_time: TimeStep) -> None:
    """Stop execution after the specified time interval.

    :param sleep_time: Time interval in seconds.
    """
    _time = sleep_time.total_seconds() if isinstance(sleep_time, timedelta) else sleep_time
    await asyncio.sleep(_time)


async def stop_keyboard(key: str = "q") -> None:
    """Stop execution if a key is pressed.

    :param key: Key to be pressed (default: "q").
    """
    while True:
        if keyboard.is_pressed(key):
            break
        await asyncio.sleep(0)


def execution_loop(
    nodes_file: Path,
    nodes_sheet: str,
    output_file: Path | None = None,
    stop_after: TimeStep | None = None,
    sub_interval: TimeStep = 1,
    write_interval: TimeStep = 1,
    eneffco_usr: str | None = None,
    eneffco_pw: str | None = None,
    eneffco_api_token: str | None = None,
    verbosity: int = 2,
) -> None:
    """Execute the subscription and publishing loop.

    :param nodes_file: Path to Excel sheet with node specification.
    :param nodes_sheet: Excel sheet name.
    :param output_file: Path to the CSV output file (optional) - One of output_file or publish_opcua is required.
    :param stop_after: Stop recording data automatically after X seconds
    :param sub_interval: Interval for subscription data.
    :param write_interval: Interval for writing to CSV file
    :param eneffco_usr: EnEffCo username.
    :param eneffco_pw: EnEffCo password.
    :param eneffco_api_token: API token for the EnEffCo connector.
    :param verbosity: Verbosity level (between 0 - no output and 4 - debug).
    """
    log.setLevel(verbosity * 10)

    nodes = Node.from_excel(nodes_file, nodes_sheet)
    connections = Connection.from_nodes(nodes, usr=eneffco_usr, pwd=eneffco_pw, api_token=eneffco_api_token)

    # Start handler
    subscription_handler = sub_handlers.MultiSubHandler()

    if output_file is None:
        raise ValueError("Specify at least an output_file")

    if output_file is not None:
        output_file = pathlib.Path(output_file)
        if output_file.is_file() or output_file.is_dir():
            try:
                pathlib.Path(output_file).unlink()
            except FileNotFoundError:
                pass

        subscription_handler.register(sub_handlers.CsvSubHandler(output_file, write_interval=write_interval))
    loop = asyncio.get_event_loop()
    task = loop.create_task(logger(10))
    background_tasks = set()
    background_tasks.add(task)

    try:
        for connection in connections.values():
            # Start connections without passing on interrupt signals
            try:
                connection.subscribe(subscription_handler, interval=sub_interval)
            except ConnectionError as e:
                log.warning(str(e))

        log.warning("Starting processing loop")
        if stop_after is not None:
            log.info(f"Process will stop after {stop_after} s.")
            loop.run_until_complete(stop_execution(stop_after))
        else:
            log.warning("Use q to stop recording data (It might take some time to react).")
            loop.run_until_complete(stop_keyboard("q"))
            log.warning("Detected key press, stopping execution.")

    finally:
        log.info("Closing connections and handlers")
        for connection in connections.values():
            connection.close_sub()

        subscription_handler.close()


if __name__ == "__main__":
    args = parse_args()

    execution_loop(
        args.nodes_file,
        args.nodes_sheet,
        args.output_file,
        args.stop_after,
        args.sub_interval,
        args.write_interval,
        args.eneffco_usr,
        args.eneffco_pw,
        args.verbosity,
    )
